from openride.core.numba.interpolation import interpolate_line_index_at_xy, linear_interpolation

import numba
import numpy as np

from typing import Tuple, Union

EPSILON = np.finfo(float).eps


@numba.njit(cache=True)
def vertices_to_distances(vertices: np.ndarray, bird_eye_view: bool = True) -> np.ndarray:
    distances_squared = (vertices[1:, 0] - vertices[:-1, 0]) ** 2 + (vertices[1:, 1] - vertices[:-1, 1]) ** 2
    if not bird_eye_view:
        distances_squared += (vertices[1:, 2] - vertices[:-1, 2]) ** 2
    distances = distances_squared**0.5
    distances = np.append(0, distances)
    return np.cumsum(distances)


@numba.njit(cache=True)
def index_to_distance(vertices: np.ndarray, index: float) -> float:
    distances = vertices_to_distances(vertices)
    if index >= vertices.shape[0] - 1:
        return distances[-1]
    return linear_interpolation(distances, index)


@numba.njit(cache=True)
def distance_to_index(vertices: np.ndarray, distance: float) -> float:
    if distance <= 0.0:
        return 0.0
    distances = vertices_to_distances(vertices)
    index = np.searchsorted(distances - distance, 0)
    if index > distances.size - 1:
        return distances.size - 1
    index += (distance - distances[index]) / (distances[index] - distances[index - 1] + EPSILON)
    return index


@numba.njit(cache=True)
def left_or_right_from_line(vertices: np.ndarray, x: float, y: float) -> int:
    i = int(interpolate_line_index_at_xy(vertices, x, y))
    i0 = max([i, 0])
    i1 = min([i + 1, vertices.shape[0] - 1])
    return np.sign(
        (vertices[i1, 0] - vertices[i0, 0]) * (y - vertices[i0, 1])
        - (vertices[i1, 1] - vertices[i0, 1]) * (x - vertices[i0, 0])
    )


@numba.njit(cache=True)
def tangent_angle(vertices: np.ndarray, index: int) -> float:
    i1 = max([0, index - 1])
    i2 = min([index + 1, vertices.shape[0] - 1])
    delta_x = vertices[i2, 0] - vertices[i1, 0]
    delta_y = vertices[i2, 1] - vertices[i1, 1]
    ratio = delta_y / (delta_x + 1e-9)
    angle = np.arctan(ratio)
    angle = angle + np.pi if delta_x < 0 else angle
    return angle % (2 * np.pi)


@numba.njit(cache=True)
def tangent_angles(vertices: np.ndarray) -> np.ndarray:
    angles = np.zeros(vertices.shape[0])
    delta_x = vertices[2:, 0] - vertices[:-2, 0]
    delta_y = vertices[2:, 1] - vertices[:-2, 1]
    ratio = delta_y / (delta_x + 1e-9)
    angles[1:-1] = np.arctan(ratio)
    angles[1:-1][np.where(delta_x < 0)] += np.pi
    angles[1:-1] = angles[1:-1] % (2 * np.pi)
    angles[0] = tangent_angle(vertices, 0)
    angles[-1] = tangent_angle(vertices, vertices.shape[0] - 1)
    return angles


@numba.njit(cache=True)
def angle_difference(angle1: float, angle2: float) -> float:
    two_pi = 2.0 * np.pi
    return min((two_pi - abs(angle1 - angle2), abs(angle1 - angle2))) % two_pi


@numba.njit(cache=True)
def angle_differences(angles1: float, angles2: float) -> float:
    two_pi = 2.0 * np.pi
    return np.array(
        [
            min((two_pi - abs(angles1[i] - angles2[i]), abs(angles1[i] - angles2[i]))) % two_pi
            for i in range(angles1.size)
        ]
    )


@numba.njit(cache=True)
def project_point_on_line(vertices: np.ndarray, x: float, y: float) -> Tuple[float, float, float]:
    """Projection is performed in bird eye view, but the coordinate interpolation is 3D."""
    index = interpolate_line_index_at_xy(vertices, x, y)
    vx = linear_interpolation(vertices[:, 0], index)
    vy = linear_interpolation(vertices[:, 1], index)
    vz = linear_interpolation(vertices[:, 2], index)
    return vx, vy, vz


@numba.njit(cache=True)
def radius_of_curvature(vertices: np.ndarray) -> np.ndarray:
    # https://stackoverflow.com/questions/61945776/curve-radius-in-python
    R = np.full(vertices.shape[0], np.inf)

    dx = vertices[1:, 0] - vertices[:-1, 0]  # First derivative of x (centered)
    dy = vertices[1:, 1] - vertices[:-1, 1]  # First derivative of y (centered)
    ddx = dx[1:] - dx[:-1]  # Second derivative of x (centered)
    ddy = dy[1:] - dy[:-1]  # Second derivative of y (centered)

    # Remove edges to match shapes
    dx = dx[:-1]
    dy = dy[:-1]

    R[:-2] = (dx**2 + dy**2) ** 1.5 / (dx * ddy - dy * ddx + EPSILON)

    # Deal with edges
    R[-2:] = R[-3]

    return R


@numba.njit(cache=True)
def apply_lateral_deviations(
    vertices: np.ndarray,
    tangent_angles: np.ndarray,
    lateral_deviations: Union[np.ndarray, float],
) -> np.ndarray:
    """Returns the route vertices after applying the lateral deviations"""
    dx = lateral_deviations * np.sin(tangent_angles)
    dy = lateral_deviations * np.cos(tangent_angles)
    shifted_vertices = np.copy(vertices)
    shifted_vertices[:, 0] -= dx
    shifted_vertices[:, 1] += dy
    return shifted_vertices


@numba.njit(cache=True)
def resample_linear(vertices: np.ndarray, N: int):
    distances = vertices_to_distances(vertices)
    new_distances = np.linspace(0, distances[-1], N)
    new_vertices = np.zeros((new_distances.size, 3))
    indices = np.zeros(new_distances.size)

    for i in range(new_distances.size):
        indices[i] = distance_to_index(vertices, new_distances[i])
        new_vertices[i, 0] = linear_interpolation(vertices[:, 0], indices[i])
        new_vertices[i, 1] = linear_interpolation(vertices[:, 1], indices[i])
        new_vertices[i, 2] = linear_interpolation(vertices[:, 2], indices[i])

    return new_vertices, indices



from openride import Point, BoundingBox, Polyline, Polygon, Size, bird_eye_view_contains
from openride.test import get_pytest_benchmark


pytest_benchmark = get_pytest_benchmark(group="operations.bird_eye_view_contains")


@pytest_benchmark
def test_bev_contains_box_point_true(benchmark):
    box = BoundingBox(Point(4, 8, -3))
    point = Point(4.5, 7.5, 5)
    assert benchmark(bird_eye_view_contains, box, point)


def test_bev_contains_box_point_false():
    assert not bird_eye_view_contains(BoundingBox(), Point(0, 5, 0))


@pytest_benchmark
def test_bev_contains_box_box_true(benchmark):
    box1 = BoundingBox()
    box2 = BoundingBox(Point(0, 0, 5), size=Size(0.5, 0.5, 2.5))
    assert benchmark(bird_eye_view_contains, box1, box2)


def test_bev_contains_box_box_false():
    assert not bird_eye_view_contains(BoundingBox(), BoundingBox(Point(0, 0, 5), size=Size(2.5, 0.5, 2.5)))


@pytest_benchmark
def test_bev_contains_box_polyline_true(benchmark):
    box = BoundingBox()
    polyline = Polyline([Point(), Point(0.5, 0.3, 2), Point(-0.3, 0.7, -2)])
    assert benchmark(bird_eye_view_contains, box, polyline)


def test_bev_contains_box_polyline_false():
    polyline = Polyline([Point(), Point(2.5, 0.3, 2), Point(-0.3, 0.7, -2)])
    assert not bird_eye_view_contains(BoundingBox(), polyline)


@pytest_benchmark
def test_bev_contains_box_polygon_true(benchmark):
    box = BoundingBox()
    polygon = Polygon([Point(), Point(0.5, 0.3, 2), Point(-0.3, 0.7, -2)])
    assert benchmark(bird_eye_view_contains, box, polygon)


def test_bev_contains_box_polygon_false():
    polygon = Polygon([Point(), Point(2.5, 0.3, 2), Point(-0.3, 0.7, -2)])
    assert not bird_eye_view_contains(BoundingBox(), polygon)


@pytest_benchmark
def test_bev_contains_polyline_point_true(benchmark):
    polyline = Polyline([Point(x, x, x) for x in range(5)])
    point = Point(1.3, 1.3, -5)
    assert benchmark(bird_eye_view_contains, polyline, point)


def test_bev_contains_polyline_point_false():
    polyline = Polyline([Point(x, x, x) for x in range(5)])
    assert not bird_eye_view_contains(polyline, Point(1.2, 1.3, -5))


@pytest_benchmark
def test_bev_contains_polyline_polyline_true(benchmark):
    pl1 = Polyline([Point(x, x, x) for x in range(5)])
    pl2 = Polyline([Point(x, x, -x) for x in range(1, 4)])
    assert benchmark(bird_eye_view_contains, pl1, pl2)


def test_bev_contains_polyline_polyline_false():
    pl1 = Polyline([Point(x, x, x) for x in range(5)])
    pl2 = Polyline([Point(x, x, -x) for x in range(1, 4)] + [Point(2, 1, 0)])
    assert not bird_eye_view_contains(pl1, pl2)


@pytest_benchmark
def test_bev_contains_polygon_point_true(benchmark):
    polygon = Polygon([Point(1, 1), Point(-1, 1), Point(-1, -1), Point(1, -1)])
    point = Point(0, 0, 4)
    assert benchmark(bird_eye_view_contains, polygon, point)


def test_bev_contains_polygon_point_false():
    polygon = Polygon([Point(1, 1), Point(-1, 1), Point(-1, -1), Point(1, -1)])
    assert not bird_eye_view_contains(polygon, Point(3, 0, 4))


@pytest_benchmark
def test_bev_contains_polygon_box_true(benchmark):
    polygon = Polygon([Point(2, 2), Point(-2, 2), Point(-2, -2), Point(2, -2)])
    box = BoundingBox(Point(0, 0, -3))
    assert benchmark(bird_eye_view_contains, polygon, box)


def test_bev_contains_polygon_box_false():
    polygon = Polygon([Point(2, 2), Point(-2, 2), Point(-2, -2), Point(2, -2)])
    assert not bird_eye_view_contains(polygon, BoundingBox(Point(0, 1.5, -3)))


@pytest_benchmark
def test_bev_contains_polygon_polyline_true(benchmark):
    polygon = Polygon([Point(3, 3), Point(-3, 3), Point(-3, -3), Point(3, -3)])
    polyline = Polyline([Point(x, x, x) for x in range(2)])
    assert benchmark(bird_eye_view_contains, polygon, polyline)


def test_bev_contains_polygon_polyline_false():
    polygon = Polygon([Point(2, 2), Point(-2, 2), Point(-2, -2), Point(2, -2)])
    polyline = Polyline([Point(x, x, x) for x in range(5)])
    assert not bird_eye_view_contains(polygon, polyline)


@pytest_benchmark
def test_bev_contains_polygon_polygon_true(benchmark):
    pg1 = Polygon([Point(3, 3), Point(-3, 3), Point(-3, -3), Point(3, -3)])
    pg2 = Polygon([Point(2, 2, 3), Point(-2, 2), Point(-2, -2), Point(2, -2)])
    assert benchmark(bird_eye_view_contains, pg1, pg2)


def test_bev_contains_polygon_polygon_false():
    pg1 = Polygon([Point(3, 3), Point(-3, 3), Point(-3, -3), Point(3, -3)])
    pg2 = Polygon([Point(2, 2, 3), Point(-2, 2), Point(-2, -2), Point(2, -2)])
    assert not bird_eye_view_contains(pg2, pg1)



import numba
import numpy as np


@numba.njit(cache=True)
def linear_interpolation(array: np.ndarray, index: float) -> float:
    if index >= array.size - 1:
        return array[-1]
    int_index = int(index)
    remainder = index % 1
    x0 = array[int_index]
    x1 = array[int_index + 1]
    return x0 * (1 - remainder) + x1 * (remainder)


@numba.njit(cache=True)
def linear_interpolation_angle(array: np.ndarray, index: float) -> float:
    if index >= array.size - 1:
        return array[-1]
    int_index = int(index)
    remainder = index % 1
    x0 = array[int_index]
    x1 = array[int_index + 1]
    if x1 - x0 > np.pi:
        x1 -= 2 * np.pi
    if x1 - x0 < -np.pi:
        x1 += 2 * np.pi
    return (x0 * (1 - remainder) + x1 * (remainder)) % (2 * np.pi)


@numba.njit(cache=True)
def interpolate_line_index_at_xy(vertices: np.ndarray, x: float, y: float, discrete_index: bool = False) -> float:
    """Return the interpolated index (float) of projected (x,y) coordinates onto a line (vertices)."""

    point = np.array([x, y])

    # Find nearest index (integer, not interpolated)
    diff = vertices[:, :2] - point
    distances_sq = diff[:, 0] ** 2 + diff[:, 1] ** 2
    index_int = np.argmin(distances_sq)

    if discrete_index:
        return index_int

    closest_point = vertices[index_int][:2]

    # Try interpolating between closest point and next point
    if index_int + 1 < len(vertices):
        next_point = vertices[index_int + 1][:2]
        delta_distance = np.sum((closest_point - next_point) ** 2)
        delta_index = np.sum((point - closest_point) * (next_point - closest_point)) / (delta_distance + 1e-16)
        if (
            delta_index <= 1 and delta_index >= 0
        ):  # This means that point is between the closest point and the next point
            return index_int + delta_index

    # Try interpolating between closest point and previous point
    if index_int >= 1:
        previous_point = vertices[index_int - 1][:2]
        delta_distance = np.sum((closest_point - previous_point) ** 2)
        delta_index = max(
            0,
            min(
                1,
                np.sum((point - closest_point) * (previous_point - closest_point)) / (delta_distance + 1e-16),
            ),
        )
        return index_int - delta_index

    return index_int



from openride import ImageViewer, BoundingBox

import numpy as np


if __name__ == "__main__":

    box = BoundingBox.from_raw(x=0, y=0, z=5, roll=0.3)

    v = ImageViewer()
    v.set_fov(resolution=(500, 500), fov_degrees=70)

    image = np.zeros((500, 500, 3)) * 255
    v.set_window_size(image.shape[:2])

    while True:

        box.rotation.pitch += 0.01

        v.draw_image(image)

        v.draw_bounding_box(box, linewidth=5, color=(0,1,0), alpha=0.5)

        v.labelize("allo", pos=box.position, color=(1,0,0), linewidth=2)

        v.draw_rectangle(20, 20, 480, 480, color=(0,0,1), linewidth=1, fill=True, alpha=0.1)

        v.update()



from openride import bird_eye_view_distance, BoundingBox, Geometry, Point, Polygon, Polyline, Rotation, Size
from openride.test import get_pytest_benchmark
from openride.test.random_core_generator import get_random

from shapely import geometry

import pytest
import numpy as np


pytest_benchmark = get_pytest_benchmark(group="operations.bird_eye_view_distance")


@pytest_benchmark
def test_bev_distance_point_point(benchmark):
    p1 = Point(0, 0, 0)
    p2 = Point(1, 1, 1)
    assert benchmark(bird_eye_view_distance, p1, p2) == 2**0.5


def test_bev_distance_point_box_inside():
    p = Point(0, 0, 0)
    b = BoundingBox()
    assert bird_eye_view_distance(p, b) == 0.0


@pytest_benchmark
def test_bev_distance_point_box(benchmark):
    p = Point(2, 0, 0)
    b = BoundingBox()
    assert benchmark(bird_eye_view_distance, p, b) == 1.0


@pytest_benchmark
def test_bev_distance_point_line(benchmark):
    p = Point(2, 2, 0)
    l = Polyline([Point(x) for x in range(5)])
    assert benchmark(bird_eye_view_distance, p, l) == 2.0


@pytest_benchmark
def test_bev_distance_point_polygon(benchmark):
    p = Point(2, 2, 0)
    poly = Polygon([Point(1, 1), Point(-1, 1), Point(-1, -1), Point(1, -1)])
    assert benchmark(bird_eye_view_distance, p, poly) == 2**0.5


@pytest_benchmark
def test_bev_distance_box_box(benchmark):
    b1 = BoundingBox(Point(0, 0, 0), Rotation(0, 0, 0), Size(1, 1, 1))
    b2 = BoundingBox(Point(2 * 2**0.5 + 1, 0, 0), Rotation(0, 0, np.pi / 4), Size(1, 1, 1))
    assert benchmark(bird_eye_view_distance, b1, b2) == pytest.approx(2**0.5)


def test_bev_distance_box_box_overlap():
    b1 = BoundingBox()
    b2 = BoundingBox(Point(0.5, 0.5))
    assert bird_eye_view_distance(b1, b2) == 0.0


@pytest_benchmark
def test_bev_distance_box_line(benchmark):
    b = BoundingBox(Point(0, 0, 0), Rotation(0, 0, 0), Size(1, 1, 1))
    l = Polyline([Point(x + 2) for x in range(10)])
    assert benchmark(bird_eye_view_distance, b, l) == pytest.approx(1.0)


def test_bev_distance_box_line_overlap():
    b = BoundingBox(Point(0, 0, 0), Rotation(0, 0, 0), Size(1, 1, 1))
    l = Polyline([Point(x) for x in range(10)])
    assert pytest.approx(bird_eye_view_distance(b, l)) == 0.0


@pytest_benchmark
def test_bev_distance_box_polygon(benchmark):
    b = BoundingBox(Point(3, 0, 0), Rotation(0, 0, 0), Size(1, 1, 1))
    p = Polygon([Point(1, 1), Point(-1, 1), Point(-1, -1), Point(1, -1)])
    assert benchmark(bird_eye_view_distance, b, p) == pytest.approx(1.0)


def test_bev_distance_box_polygon_overlap():
    b = BoundingBox(Point(0, 0, 0), Rotation(0, 0, 0), Size(1, 1, 1))
    p = Polygon([Point(1, 1), Point(-1, 1), Point(-1, -1), Point(1, -1)])
    assert pytest.approx(bird_eye_view_distance(b, p)) == 0.0


@pytest_benchmark
def test_bev_distance_line_point(benchmark):
    p = Point(5.3, 5, 5)
    l = Polyline([Point(x) for x in range(10)])
    assert benchmark(bird_eye_view_distance, l, p) == pytest.approx(5.0)


def test_bev_distance_line_point_overlap():
    p = Point(5.3, 0, 5)
    l = Polyline([Point(x) for x in range(10)])
    assert pytest.approx(bird_eye_view_distance(l, p)) == 0.0


@pytest_benchmark
def test_bev_distance_line_line(benchmark):
    l1 = Polyline([Point(x) for x in range(10)])
    l2 = Polyline([Point(5.5, y + 1) for y in range(10)])
    assert benchmark(bird_eye_view_distance, l1, l2) == 1.0


def test_bev_distance_line_line_overlap():
    l1 = Polyline([Point(x) for x in range(10)])
    l2 = Polyline([Point(5, y - 5) for y in range(10)])
    assert bird_eye_view_distance(l1, l2) == 0.0


def test_bev_distance_line_polygon_overlap():
    l = Polyline([Point(x) for x in range(10)])
    p = Polygon([Point(1, 1), Point(-1, 1), Point(-1, -1), Point(1, -1)])
    assert bird_eye_view_distance(l, p) == 0.0


@pytest_benchmark
def test_bev_distance_polygon_polygon(benchmark):
    p1 = Polygon([Point(1, 1), Point(-1, 1), Point(-1, -1), Point(1, -1)])
    p2 = Polygon([Point(2, 0), Point(3, -1), Point(4, 0), Point(3, 1)])
    assert benchmark(bird_eye_view_distance, p1, p2) == pytest.approx(1.0)


def test_bev_distance_line_line_same_as_shapely():
    for _ in range(10):
        l1 = get_random(Polyline)
        l2 = get_random(Polyline)
        assert pytest.approx(bird_eye_view_distance(l1, l2)) == l1.to_shapely().distance(l2.to_shapely())


def test_bev_distance_point_point_same_as_shapely():
    for _ in range(10):
        p1 = get_random(Point)
        p2 = get_random(Point)
        assert pytest.approx(bird_eye_view_distance(p1, p2)) == p1.to_shapely().distance(p2.to_shapely())


def test_bev_distance_box_point_same_as_shapely():
    for _ in range(10):
        p = get_random(Point)
        b = get_random(BoundingBox)
        assert pytest.approx(bird_eye_view_distance(b, p)) == b.to_shapely().distance(p.to_shapely())


def test_bev_distance_box_box_same_as_shapely():
    for _ in range(10):
        b1 = get_random(BoundingBox)
        b2 = get_random(BoundingBox)
        assert pytest.approx(bird_eye_view_distance(b1, b2)) == b1.to_shapely().distance(b2.to_shapely())


def test_bev_distance_box_line_same_as_shapely():
    for _ in range(10):
        l = get_random(Polyline)
        b = get_random(BoundingBox)
        assert pytest.approx(bird_eye_view_distance(b, l)) == b.to_shapely().distance(l.to_shapely())


def test_bev_distance_line_point_same_as_shapely():
    for _ in range(10):
        p = get_random(Point)
        l = get_random(Polyline)
        assert pytest.approx(bird_eye_view_distance(l, p)) == l.to_shapely().distance(p.to_shapely())


def test_bev_distance_line_box_same_as_shapely():
    for _ in range(10):
        l = get_random(Polyline)
        b = get_random(BoundingBox)
        assert pytest.approx(bird_eye_view_distance(l, b)) == l.to_shapely().distance(b.to_shapely())


def test_bev_distance_polygon_point_same_as_shapely():
    for _ in range(10):
        poly = get_random(Polygon)
        p = get_random(Point)
        assert pytest.approx(bird_eye_view_distance(poly, p)) == poly.to_shapely().distance(p.to_shapely())


def test_bev_distance_polygon_box_same_as_shapely():
    for _ in range(10):
        p = get_random(Polygon)
        b = get_random(BoundingBox)
        assert pytest.approx(bird_eye_view_distance(p, b)) == p.to_shapely().distance(b.to_shapely())


def test_bev_distance_polygon_line_same_as_shapely():
    for _ in range(10):
        p = get_random(Polygon)
        l = get_random(Polyline)
        assert pytest.approx(bird_eye_view_distance(p, l)) == p.to_shapely().distance(l.to_shapely())


def test_bev_distance_polygons_same_as_shapely():
    for _ in range(10):
        p1 = get_random(Polygon)
        p2 = get_random(Polygon)
        assert pytest.approx(bird_eye_view_distance(p1, p2)) == p1.to_shapely().distance(p2.to_shapely())


def test_default_geometry_distance_from_shapely():
    class MockGeometry(Geometry):
        def to_shapely(self):
            return geometry.Point([1, 1, 1])

        def transform(self, transform):
            pass

    p1 = Point(0, 0, 0)
    p2 = MockGeometry()
    assert bird_eye_view_distance(p1, p2) == 2**0.5



from openride.viewer.model_loader import ModelLoader

import numpy as np
import vtk
import vtk.util.numpy_support as vtk_np

from typing import List


def get_point_actor(
    x: float,
    y: float,
    z: float,
    color: tuple = (1, 1, 1),
    size: int = 0.15,
    alpha: float = 1,
    resolution: int = 10,
) -> vtk.vtkActor:
    source = vtk.vtkSphereSource()
    source.SetThetaResolution(resolution)
    source.SetPhiResolution(resolution)
    source.SetCenter(x, y, z)
    source.SetRadius(size)
    return source_to_actor(source, color, alpha)


BOX_ACTORS_CACHE = {}


def get_box_actor(
    x: float,
    y: float,
    z: float,
    sx: float,
    sy: float,
    sz: float,
    roll: float,
    pitch: float,
    yaw: float,
    color: tuple = None,
    alpha: float = 1,
) -> vtk.vtkActor:

    # Creating these actors is expansive, so we reuse them if an identical one has already been created.
    args = (color, alpha)
    if args not in BOX_ACTORS_CACHE:
        source = vtk.vtkCubeSource()
        BOX_ACTORS_CACHE[args] = source_to_actor(source, color, alpha)

    actor = vtk.vtkActor()
    actor.ShallowCopy(BOX_ACTORS_CACHE[args])

    actor.SetScale(2 * sx, 2 * sy, 2 * sz)
    rotate_actor(actor, roll, pitch, yaw)
    move_actor(actor, x, y, z)
    return actor


def get_line_actor(
    vertices: np.ndarray,
    color: tuple = (1, 1, 1),
    alpha: float = 1,
    linewidth: int = 2,
    dashed: bool = False,
) -> vtk.vtkActor:
    polydata = vtk.vtkPolyData()
    actor = polydata_to_actor(polydata, color, alpha)
    actor.GetProperty().SetLineWidth(linewidth)
    points = vtk.vtkPoints()
    points.SetData(vtk_np.numpy_to_vtk(vertices))
    polydata.SetPoints(points)

    lines = vtk.vtkCellArray()
    step = 1 if not dashed else 2
    for i in range(0, vertices.shape[0] - 1, step):
        lines.InsertNextCell(2)
        lines.InsertCellPoint(i)
        lines.InsertCellPoint(i + 1)
    polydata.SetLines(lines)
    return actor


def get_lines_actor(
    list_vertices: List[np.ndarray],
    color: tuple = (1, 1, 1),
    alpha: float = 1,
    linewidth: int = 2,
) -> vtk.vtkActor:
    polydata = vtk.vtkPolyData()
    actor = polydata_to_actor(polydata, color, alpha)
    actor.GetProperty().SetLineWidth(linewidth)
    points = vtk.vtkPoints()
    points.SetData(vtk_np.numpy_to_vtk(np.vstack(list_vertices)))
    polydata.SetPoints(points)

    lines = vtk.vtkCellArray()
    i = 0
    for vertices in list_vertices:
        for _ in range(0, vertices.shape[0] - 1, 1):
            lines.InsertNextCell(2)
            lines.InsertCellPoint(i)
            lines.InsertCellPoint(i + 1)
            i += 1
        i += 1
    polydata.SetLines(lines)
    return actor


def get_polygon_polydata(vertices: np.ndarray) -> vtk.vtkPolyData:

    points = vtk.vtkPoints()
    polygon = vtk.vtkPolygon()
    polygon.GetPointIds().SetNumberOfIds(vertices.shape[0])

    for i, p in enumerate(vertices):
        polygon.GetPointIds().SetId(i, i)
        points.InsertNextPoint(p[0], p[1], p[2])

    # Add the polygon to a list of polygons
    polygons = vtk.vtkCellArray()
    polygons.InsertNextCell(polygon)

    # Create a PolyData
    polygonPolyData = vtk.vtkPolyData()
    polygonPolyData.SetPoints(points)
    polygonPolyData.SetPolys(polygons)

    pdflt = vtk.vtkTriangleFilter()
    pdflt.SetInputData(polygonPolyData)
    pdflt.Update()
    polygonPolyData = pdflt.GetOutput()

    return polygonPolyData


def get_polygon_actor(vertices: np.ndarray, color: tuple = (1, 1, 1), alpha: float = 1) -> vtk.vtkActor:
    return polydata_to_actor(get_polygon_polydata(vertices), color, alpha)


def get_polygons_actor(vertices_list: List[np.ndarray], color: tuple = (1, 1, 1), alpha: float = 1) -> vtk.vtkActor:

    appendFilter = vtk.vtkAppendPolyData()
    for vertices in vertices_list:
        appendFilter.AddInputData(get_polygon_polydata(vertices))
    appendFilter.Update()

    pdflt = vtk.vtkTriangleFilter()
    pdflt.SetInputData(appendFilter.GetOutput())
    pdflt.Update()
    polygonPolyData = pdflt.GetOutput()

    return polydata_to_actor(polygonPolyData, color, alpha)


def get_point_cloud_actor(
    vertices: np.ndarray,
    colors: np.ndarray = None,
    color: tuple = None,
    size: int = 10,
    alpha: float = 1,
):
    polydata = vtk.vtkPolyData()
    actor = polydata_to_actor(polydata, alpha=alpha)
    actor.GetProperty().SetPointSize(size)
    actor.GetProperty().SetRenderPointsAsSpheres(1)
    points = vtk.vtkPoints()
    cells = vtk.vtkCellArray()
    N = vertices.shape[0]
    points.SetData(vtk_np.numpy_to_vtk(vertices))
    cells_npy = np.vstack([np.ones(N, dtype=np.int64), np.arange(N, dtype=np.int64)]).T.flatten()
    cells.SetCells(N, vtk_np.numpy_to_vtkIdTypeArray(cells_npy))
    polydata.SetPoints(points)
    polydata.SetVerts(cells)

    # Colorize point cloud
    # FIXME: This is slow for large point clouds because of the loop. But a better solution is not trivial.
    if color is None:
        vtkcolors = vtk.vtkUnsignedCharArray()
        vtkcolors.SetNumberOfComponents(3)
        for i, p in enumerate(vertices):
            vtkcolors.InsertNextTuple(colors[i])
        polydata.GetCellData().SetScalars(vtkcolors)
    elif isinstance(color, tuple):
        actor.GetProperty().SetColor(*color)
    else:
        # (N,3) rgb color array
        vtkcolors = vtk.vtkUnsignedCharArray()
        vtkcolors.SetNumberOfComponents(3)
        for i, p in enumerate(vertices):
            vtkcolors.InsertNextTuple(color[i] * 255)
        polydata.GetCellData().SetScalars(vtkcolors)

    return actor


# Creating these actors is expensive, so we initialize a bunch of them once and reuse them
TEXT_ACTORS_CACHE = [vtk.vtkCaptionActor2D() for _ in range(100)]


def get_text_actor(
    text: str,
    x: float,
    y: float,
    z: float,
    fontsize: int = 16,
    _text_actor_index: int = None,
) -> vtk.vtkCaptionActor2D:

    if _text_actor_index is None:
        actor = vtk.vtkCaptionActor2D()
    else:
        # Because vtkCaptionActor2D are expensive to initialize, we reuse the same actors stored in a list that we expand if needed
        while _text_actor_index >= len(TEXT_ACTORS_CACHE):
            TEXT_ACTORS_CACHE.append(vtk.vtkCaptionActor2D())
        actor = TEXT_ACTORS_CACHE[_text_actor_index]

    actor.SetCaption(text)
    actor.SetAttachmentPoint(x, y, z)
    actor.BorderOff()
    actor.GetTextActor().SetTextScaleModeToNone()
    actor.GetCaptionTextProperty().SetFontSize(fontsize)
    actor.GetCaptionTextProperty().SetShadow(0)
    return actor


def get_cone_actor(
    x1: float,
    y1: float,
    z1: float,
    x2: float,
    y2: float,
    z2: float,
    color: tuple = (1, 1, 1),
    alpha: float = 1,
    headsize: float = 1,
):
    cone_source = vtk.vtkConeSource()
    cone_source.SetHeight(2 * headsize)
    cone_source.SetRadius(headsize)
    cone_source.SetCenter(x2, y2, z2)
    cone_source.SetDirection(x2 - x1, y2 - y1, z2 - z1)
    return source_to_actor(cone_source, color, alpha)


MODEL_ACTORS_CACHE = {}


def get_model_actor(
    loader: ModelLoader,
    file: str,
    number_polygons: int,
    x: float,
    y: float,
    z: float,
    sx: float,
    sy: float,
    sz: float,
    roll: float,
    pitch: float,
    yaw: float,
    color: tuple = (1, 1, 1),
    alpha: float = 1,
) -> vtk.vtkActor:

    # Creating these actors is expansive, so we reuse them if an identical one has already been created.
    args = (file, color, alpha, number_polygons)
    if args not in MODEL_ACTORS_CACHE:
        polydata = loader(file, n_polygons=number_polygons)
        _actor = polydata_to_actor(polydata, color, alpha)
        MODEL_ACTORS_CACHE[args] = _actor

    actor = vtk.vtkActor()
    actor.ShallowCopy(MODEL_ACTORS_CACHE[args])

    x0, x1, y0, y1, z0, z1 = actor.GetBounds()
    actor.SetScale(2 * sx / (x1 - x0), 2 * sy / (y1 - y0), 2 * sz / (z1 - z0))
    rotate_actor(actor, roll, pitch, yaw)
    x0, x1, y0, y1, z0, z1 = actor.GetBounds()
    move_actor(actor, x - (x0 + x1) / 2, y - (y0 + y1) / 2, z - (z0 + z1) / 2)

    return actor


def mapper_to_actor(mapper: vtk.vtkPolyDataMapper, color: tuple = (1, 1, 1), alpha: float = 1.0) -> vtk.vtkActor:
    actor = vtk.vtkActor()
    actor.SetMapper(mapper)
    actor.GetProperty().SetColor(*color)
    actor.GetProperty().SetOpacity(alpha)
    return actor


def source_to_actor(source, color: tuple = (1, 1, 1), alpha: float = 1.0) -> vtk.vtkActor:
    mapper = vtk.vtkPolyDataMapper()
    mapper.SetInputConnection(source.GetOutputPort())
    return mapper_to_actor(mapper, color, alpha)


def polydata_to_actor(polydata, color: tuple = (1, 1, 1), alpha: float = 1.0) -> vtk.vtkActor:
    mapper = vtk.vtkPolyDataMapper()
    mapper.SetInputData(polydata)
    return mapper_to_actor(mapper, color, alpha)


def array_to_texture(array: np.ndarray) -> vtk.vtkTexture:
    grid = vtk.vtkImageData()
    grid.SetDimensions(array.shape[0], array.shape[1], 1)
    vtkarr = vtk_np.numpy_to_vtk(array.swapaxes(0, 1).reshape((-1, array.shape[-1]), order="F"))
    vtkarr.SetName("Image")
    grid.GetPointData().AddArray(vtkarr)
    grid.GetPointData().SetActiveScalars("Image")
    vtex = vtk.vtkTexture()
    vtex.SetColorModeToDirectScalars()
    vtex.SetInputDataObject(grid)
    return vtex


def move_actor(actor: vtk.vtkActor, x: float, y: float, z: float):
    actor.AddPosition(x, y, z)


def rotate_actor(actor: vtk.vtkActor, roll: float, pitch: float, yaw: float):
    actor.AddOrientation(np.rad2deg(roll), np.rad2deg(pitch), np.rad2deg(yaw))


def set_scale_actor(actor: vtk.vtkActor, sx: float, sy: float, sz: float):
    actor.SetScale(sx, sy, sz)



from openride.core.transform import Transform

from abc import ABC, abstractclassmethod
from typing import Any


class Collection(ABC):
    @abstractclassmethod
    def __len__(self) -> int:
        """Amount of objects in the collection"""

    @abstractclassmethod
    def append(self, obj: Any):
        """Add an object to the collection"""

    @abstractclassmethod
    def extend(self, collection: "Collection"):
        """Add another collection to this collection"""

    @abstractclassmethod
    def pop(self, index: int) -> Any:
        """Remove and return the object at index"""

    def __repr__(self):
        s = f"{self.__class__.__name__}: (N={len(self)}) \n"
        for i, obj in enumerate(self):
            s += f"{i}: {obj} \n"
        return s[:-2]

    @abstractclassmethod
    def __getitem__(self, index: int) -> Any:
        """Return the object at index"""

    @abstractclassmethod
    def transform(self, transform: Transform) -> "Collection":
        """Return a transformed copy of this collection"""



from typing import Tuple

import numba
import numpy as np


@numba.njit(cache=True)
def rotation_matrix(roll: float, pitch: float, yaw: float) -> np.ndarray:
    s1 = np.sin(roll)
    c1 = np.cos(roll)
    s2 = np.sin(pitch)
    c2 = np.cos(pitch)
    s3 = np.sin(yaw)
    c3 = np.cos(yaw)

    return np.array(
        [
            [c2 * c3, s1 * s2 * c3 - c1 * s3, s1 * s3 + c1 * s2 * c3],
            [c2 * s3, c1 * c3 + s1 * s2 * s3, c1 * s2 * s3 - s1 * c3],
            [-s2, s1 * c2, c1 * c2],
        ]
    )


@numba.njit(cache=True)
def transform_matrix(x: float, y: float, z: float, roll: float, pitch: float, yaw: float) -> np.ndarray:
    matrix = np.eye(4)
    matrix[:3, :3] = rotation_matrix(roll, pitch, yaw)
    matrix[:3, 3] = np.array((x, y, z))
    return matrix


@numba.njit(cache=True)
def transform_inverse_matrix(x: float, y: float, z: float, roll: float, pitch: float, yaw: float) -> np.ndarray:
    tf = transform_matrix(x, y, z, roll, pitch, yaw)
    inverse_matrix = np.zeros_like(tf)
    inverse_matrix[:3, :3] = tf[:3, :3].T
    inverse_matrix[:3, 3] = np.dot(-tf[:3, :3].T, tf[:3, 3])
    inverse_matrix[3, 3] = 1
    return inverse_matrix


@numba.njit(cache=True)
def euler_angles(matrix: np.ndarray) -> Tuple[float, float, float]:

    sy = np.sqrt(matrix[0, 0] * matrix[0, 0] + matrix[1, 0] * matrix[1, 0])
    singular = sy < 1e-6

    if not singular:
        x = np.arctan2(matrix[2, 1], matrix[2, 2])
        y = np.arctan2(-matrix[2, 0], sy)
        z = np.arctan2(matrix[1, 0], matrix[0, 0])
    else:
        x = np.arctan2(-matrix[1, 2], matrix[1, 1])
        y = np.arctan2(-matrix[2, 0], sy)
        z = 0

    return x, y, z


@numba.njit(cache=True)
def transform_point(x: float, y: float, z: float, matrix: np.ndarray) -> np.ndarray:
    v = np.dot(matrix, np.array([x, y, z, 1.0]))
    return v[0:3] / v[3]


@numba.njit(cache=True)
def transform_vertices(vertices: np.ndarray, matrix: np.ndarray) -> np.ndarray:
    return (np.dot(matrix[:3, :3], vertices.T) + np.expand_dims(matrix[:3, 3], -1)).T



from openride import Point, BoundingBox

from typing import Tuple

import numpy as np


class Camera:
    def __init__(self, viewer):
        self.viewer = viewer
        self._camera = viewer.renderer.GetActiveCamera()
        self._camera.SetViewUp(0, 0, 1)
        self.set_clipping_range()

        self.set_position(Point(-10, 0, 5))
        self.set_focus(Point(0, 0, 0))

    def set_position(self, position: Point):
        self._camera.SetPosition(position.x, position.y, position.z)

    def set_focus(self, focus: Point):
        self._camera.SetFocalPoint(focus.x, focus.y, focus.z)

    def get_position(self) -> Point:
        return Point(*self._camera.GetPosition())

    def get_focus(self) -> Point:
        return Point(*self._camera.GetFocalPoint())
    
    def set_position_xyz(self, x: float, y: float, z: float):
        self._camera.SetPosition(x, y, z)

    def set_focus_xyz(self, x: float, y: float, z: float):
        self._camera.SetFocalPoint(x, y, z)

    def get_position_xyz(self) -> Tuple[float, float, float]:
        return tuple(self._camera.GetPosition())

    def get_focus_xyz(self) -> Tuple[float, float, float]:
        return tuple(self._camera.GetFocalPoint())
    
    def set_clipping_range(self, range:float=100000):
        self._camera.SetClippingRange(0.1, range)

    def follow(
        self,
        bounding_box: BoundingBox,
        distance: float = 15,
        height: float = 3,
        spring: float = 0.0,
        offset: float = 0.0,
    ):
        _offset = Point(
            x = offset * np.cos(bounding_box.rotation.yaw),
            y = offset * np.sin(bounding_box.rotation.yaw),
        )
        self.set_focus(bounding_box.position + _offset)

        x = -distance * np.cos(bounding_box.rotation.yaw)
        y = -distance * np.sin(bounding_box.rotation.yaw)
        z = height

        target_position = bounding_box.position + Point(x, y, z) + _offset
        self.set_position(spring * self.get_position() + (1 - spring) * target_position)

    def is_in_front(self, point: Point) -> bool:
        position = self.get_position()
        focus = self.get_focus()
        v1 = np.array([point.x - position.x, point.y - position.y, point.z - position.z])
        v2 = np.array([focus.x - position.x, focus.y - position.y, focus.z - position.z])
        return np.dot(v1, v2) >= 0



from openride import AsciiViewer, BoundingBox, Point


if __name__ == "__main__":

    box = BoundingBox()

    v = AsciiViewer()

    v.camera.set_position(Point(-8, 0, 2))

    while True:

        box.rotation.yaw += 0.01

        v.draw_bounding_box(box)

        v.update()



from openride import Point, BoundingBox, Polyline, Polygon, bird_eye_view_distance
from openride.core.numba import (
    polygon_contains_point,
    polygon_contains_all_vertices,
    bev_distance_between_box_and_point,
)

from multipledispatch import dispatch


@dispatch(BoundingBox, Point)
def bird_eye_view_contains(obj1: BoundingBox, obj2: Point) -> bool:
    return (
        bev_distance_between_box_and_point(
            obj1.position.x,
            obj1.position.y,
            obj1.size.x,
            obj1.size.y,
            obj1.rotation.yaw,
            obj2.x,
            obj2.y,
        )
        == 0.0
    )


@dispatch(BoundingBox, BoundingBox)
def bird_eye_view_contains(obj1: BoundingBox, obj2: BoundingBox) -> bool:
    return polygon_contains_all_vertices(obj1.get_bird_eye_view_vertices(), obj2.get_bird_eye_view_vertices())


@dispatch(BoundingBox, Polyline)
def bird_eye_view_contains(obj1: BoundingBox, obj2: Polyline) -> bool:
    return polygon_contains_all_vertices(obj1.get_bird_eye_view_vertices(), obj2.vertices)


@dispatch(BoundingBox, Polygon)
def bird_eye_view_contains(obj1: BoundingBox, obj2: Polygon) -> bool:
    return polygon_contains_all_vertices(obj1.get_bird_eye_view_vertices(), obj2.vertices)


@dispatch(Polyline, Point)
def bird_eye_view_contains(obj1: Polyline, obj2: Point) -> bool:
    return bird_eye_view_distance(obj1, obj2) == 0.0


@dispatch(Polyline, Polyline)
def bird_eye_view_contains(obj1: Polyline, obj2: Polyline) -> bool:
    return all([bird_eye_view_contains(obj1, point) for point in obj2])


@dispatch(Polygon, Point)
def bird_eye_view_contains(obj1: Polygon, obj2: Point) -> bool:
    return polygon_contains_point(obj1.vertices, obj2.x, obj2.y)


@dispatch(Polygon, BoundingBox)
def bird_eye_view_contains(obj1: Polygon, obj2: BoundingBox) -> bool:
    return polygon_contains_all_vertices(obj1.vertices, obj2.get_bird_eye_view_vertices())


@dispatch(Polygon, Polyline)
def bird_eye_view_contains(obj1: Polygon, obj2: Polyline) -> bool:
    return polygon_contains_all_vertices(obj1.vertices, obj2.vertices)


@dispatch(Polygon, Polygon)
def bird_eye_view_contains(obj1: Polygon, obj2: Polygon) -> bool:
    return polygon_contains_all_vertices(obj1.vertices, obj2.vertices)



from openride.core.bounding_box import BoundingBox
from openride.core.point import Point
from openride.core.polyline import Polyline

from openride.core.numba import (
    distance_between_points_3d,
    distance_between_box_and_point_3d,
    distance_between_boxes_3d,
    distance_between_line_and_point_3d,
    distance_between_lines_3d,
    distance_between_line_and_box_3d,
)

from multipledispatch import dispatch


@dispatch(Point, Point)
def distance(obj1: Point, obj2: Point) -> float:
    return distance_between_points_3d(obj1.x, obj1.y, obj1.z, obj2.x, obj2.y, obj2.z)


@dispatch(Point, BoundingBox)
def distance(obj1: Point, obj2: BoundingBox) -> float:
    return distance_between_box_and_point_3d(
        obj2.get_vertices(),
        obj2.position.x,
        obj2.position.y,
        obj2.position.z,
        obj1.x,
        obj1.y,
        obj1.z,
    )


@dispatch(Point, Polyline)
def distance(obj1: Point, obj2: Polyline) -> float:
    return distance_between_line_and_point_3d(obj2.vertices, obj1.x, obj1.y, obj1.z)


@dispatch(BoundingBox, Point)
def distance(obj1: BoundingBox, obj2: Point) -> float:
    return distance_between_box_and_point_3d(
        obj1.get_vertices(),
        obj1.position.x,
        obj1.position.y,
        obj1.position.z,
        obj2.x,
        obj2.y,
        obj2.z,
    )


@dispatch(BoundingBox, BoundingBox)
def distance(obj1: BoundingBox, obj2: BoundingBox) -> float:
    return distance_between_boxes_3d(obj1.get_vertices(), obj2.get_vertices())


@dispatch(BoundingBox, Polyline)
def distance(obj1: BoundingBox, obj2: Polyline) -> float:
    return distance_between_line_and_box_3d(obj2.vertices, obj1.get_vertices())


@dispatch(Polyline, Point)
def distance(obj1: Polyline, obj2: Point) -> float:
    return distance_between_line_and_point_3d(obj1.vertices, obj2.x, obj2.y, obj2.z)


@dispatch(Polyline, BoundingBox)
def distance(obj1: Polyline, obj2: BoundingBox) -> float:
    return distance_between_line_and_box_3d(obj1.vertices, obj2.get_vertices())


@dispatch(Polyline, Polyline)
def distance(obj1: Polyline, obj2: Polyline) -> float:
    return distance_between_lines_3d(obj1.vertices, obj2.vertices)



from openride.viewer.models import MODEL_ROTATIONS
from openride import Rotation

from abc import ABC, abstractclassmethod

import os
import numpy as np
import vtk


class BaseModelLoader(ABC):
    @abstractclassmethod
    def __call__(self, file: str, n_polygons: int = 100000) -> vtk.vtkPolyData:
        """Loads files into vtk.vtkPolyData instances"""


def decimate(polydata:vtk.vtkPolyData, n_polygons:int) -> vtk.vtkPolyData:
    decimate = vtk.vtkQuadricDecimation()
    decimate.SetInputData(polydata)
    decimate.SetTargetReduction(1.0 - n_polygons / polydata.GetNumberOfPolys())
    decimate.Update()
    return decimate.GetOutput()


def rotate(polydata:vtk.vtkPolyData, roll:float, pitch:float, yaw:float) -> vtk.vtkPolyData:
    transform = vtk.vtkTransform()
    transform.RotateX(roll)
    transform.RotateY(pitch)
    transform.RotateZ(yaw)
    fil = vtk.vtkTransformPolyDataFilter()
    fil.SetTransform(transform)
    fil.SetInputDataObject(polydata)
    fil.Update()
    return fil.GetOutput()


class STLLoader(BaseModelLoader):
    def __init__(self):
        self._loaded_files = {}

    def __call__(self, file: str, n_polygons: int = 100000) -> vtk.vtkPolyData:
        """Loads .stl files into vtk.vtkPolyData instances"""
        if file not in self._loaded_files:
            reader = vtk.vtkSTLReader()
            reader.SetFileName(file)
            reader.Update()

            polydata = reader.GetOutput()
            polydata = decimate(polydata, n_polygons)
            rotation = MODEL_ROTATIONS.get(os.path.split(file)[-1], Rotation())
            polydata = rotate(polydata, np.rad2deg(rotation.roll), np.rad2deg(rotation.pitch), np.rad2deg(rotation.yaw))

            self._loaded_files[file] = polydata

        return self._loaded_files[file]


class OBJLoader(BaseModelLoader):
    def __init__(self):
        self._loaded_files = {}

    def __call__(self, file: str, n_polygons: int = 100000) -> vtk.vtkPolyData:
        """Loads .stl files into vtk.vtkPolyData instances"""
        if file not in self._loaded_files:
            reader = vtk.vtkOBJReader()
            reader.SetFileName(file)
            reader.Update()
            polydata = reader.GetOutput()

            polydata = reader.GetOutput()
            polydata = decimate(polydata, n_polygons)
            rotation = MODEL_ROTATIONS.get(os.path.split(file)[-1], Rotation())
            polydata = rotate(polydata, np.rad2deg(rotation.roll), np.rad2deg(rotation.pitch), np.rad2deg(rotation.yaw))

        return self._loaded_files[file]


class ModelLoader(BaseModelLoader):
    def __init__(self):
        self.loaders = {
            "stl": STLLoader(),
            "obj": OBJLoader(),
        }

    def __call__(self, file: str, n_polygons: int = 100000) -> vtk.vtkPolyData:
        loader = self.loaders.get(file.split(".")[-1])
        if loader is None:
            raise NotImplementedError
        else:
            return loader(file, n_polygons)



from openride import BoundingBox, Point, Rotation, Size, Transform
from openride.test import get_pytest_benchmark

from shapely import geometry

import numpy as np
import pytest


pytest_benchmark = get_pytest_benchmark(group="core.bounding_box")


def test_init():
    b = BoundingBox()
    assert b.position == Point()
    assert b.rotation == Rotation()
    assert b.size == Size()


def test_init_from_raw():
    b1 = BoundingBox.from_raw()
    b2 = BoundingBox()
    assert b1 == b2


@pytest_benchmark
def test_box_get_bird_eye_view_vertices(benchmark):
    box = BoundingBox()
    vertices = benchmark(box.get_bird_eye_view_vertices)
    assert vertices.shape == (4, 2)


def test_bird_eye_view_vertices_translation():
    v = BoundingBox(Point(1, 2, 3)).get_bird_eye_view_vertices()
    assert np.all(v == np.array([[2, 3], [2, 1], [0, 1], [0, 3]]))


# def test_bird_eye_view_vertices_rotation():
#     v = BoundingBox(rotation=Rotation(0, 0, np.pi / 6)).get_bird_eye_view_vertices()
#     x = np.sin(np.deg2rad(15)) * 2**0.5
#     assert np.all(pytest.approx(v) == np.array([[x, 1 + x], [1 + x, -x], [-x, -1 - x], [-1 - x, x]]))


def test_bird_eye_view_vertices_scale():
    v = BoundingBox(size=Size(2, 2, 2)).get_bird_eye_view_vertices()
    assert np.all(v == np.array([[2, 2], [2, -2], [-2, -2], [-2, 2]]))


@pytest_benchmark
def test_box_get_vertices(benchmark):
    box = BoundingBox()
    vertices = benchmark(box.get_vertices)
    assert vertices.shape == (8, 3)


def test_bounding_box_to_shapely():
    b = BoundingBox()
    assert isinstance(b.to_shapely(), geometry.Polygon)


def test_get_transform():
    b = BoundingBox()
    assert isinstance(b.get_transform(), Transform)


def test_box_transform_identity():
    b = BoundingBox()
    assert b == b.transform(Transform())


def test_box_rotation():
    b = BoundingBox()
    tf = Transform(rotation=Rotation(0, 0, np.pi / 6))
    assert pytest.approx(b.transform(tf).rotation.yaw) == np.pi / 6


@pytest_benchmark
def test_box_transform(benchmark):
    box = BoundingBox(size=Size(2, 1, 1))
    tf = Transform(Point(1, 0, 0), Rotation(0, 0, np.pi / 2))
    box_tf: BoundingBox = benchmark(box.transform, tf)
    assert pytest.approx(box_tf.position.x) == 1.0
    assert pytest.approx(box_tf.position.y) == 0.0
    assert pytest.approx(box_tf.position.z) == 0.0
    assert pytest.approx(box_tf.rotation.roll) == 0.0
    assert pytest.approx(box_tf.rotation.pitch) == 0.0
    assert pytest.approx(box_tf.rotation.yaw) == np.pi / 2



from typing import Optional, Tuple
from openride.viewer.viewer import get_last_rendered_image
from openride import Viewer
import cv2
import skvideo.io


class Recorder:
    def __init__(self, viewer: Viewer, output_file: str, fps: int = 60, resolution: Optional[Tuple[int, int]] = None):

        self.viewer = viewer
        self.viewer.set_callback(self.update)
        self.resolution = resolution

        self.video_writer = skvideo.io.FFmpegWriter(
            output_file, inputdict={"-r": str(fps)}, outputdict={"-r": str(fps)}
        )

    def update(self):

        image = get_last_rendered_image(self.viewer)

        if not self.resolution:
            self.resolution = (image.shape[0], image.shape[1])

        if (image.shape[0], image.shape[1]) != self.resolution:
            image = cv2.resize(image, self.resolution)

        self.video_writer.writeFrame(image)

    def close(self):
        self.video_writer.close()



from openride import Polyline, Point, Transform
from openride.core.numba import spherical_to_cartesian
from openride.test import get_pytest_benchmark
from openride.test.random_core_generator import get_random

from shapely import geometry

import pytest
import numpy as np


pytest_benchmark = get_pytest_benchmark(group="core.polyline")


def test_init_numpy_array():
    pl = Polyline(np.random.random((10, 3)))
    assert pl.vertices.shape == (10, 3)


def test_init_list_points():
    pl = Polyline([Point(x) for x in range(2)])
    assert np.all(pl.vertices == np.array([[0, 0, 0], [1, 0, 0]]))


def test_polyline_to_shapely():
    pl = Polyline(np.random.random((10, 3)))
    assert isinstance(pl.to_shapely(), geometry.LineString)


@pytest_benchmark
def test_line_transform_identity(benchmark):
    line = get_random(Polyline)
    tf = Transform()
    line_tf = benchmark(line.transform, tf)
    assert np.all(line.vertices == line_tf.vertices)


def test_polyline_to_shapely_single_point():
    pl = Polyline(np.random.random((1, 3)))
    assert isinstance(pl.to_shapely(), geometry.Point)


def test_len_polyline():
    pl = Polyline(np.random.random((10, 3)))
    assert len(pl) == 10


def test_append():
    pl = Polyline(np.random.random((10, 3)))
    pl.append(Point())
    assert len(pl) == 11
    assert pl[10] == Point()


def test_extent():
    pl1 = Polyline(np.random.random((10, 3)))
    pl2 = Polyline(np.random.random((10, 3)))
    pl1.extent(pl2)
    assert isinstance(pl1, Polyline)
    assert len(pl1) == 20


@pytest_benchmark
def test_polyline_get_distances(benchmark):
    pl = Polyline([Point(x) for x in range(5)])
    distances = benchmark(pl.get_distances)
    assert np.all(distances == np.arange(5))


def test_polyline_get_total_distance():
    pl = Polyline([Point(x) for x in range(5)])
    assert pl.get_total_distance() == 4


def test_polyline_get_total_distance_zero():
    assert Polyline().get_total_distance() == 0


@pytest_benchmark
def test_index_to_distance(benchmark):
    pl = Polyline([Point(x) for x in range(10)])
    distance = benchmark(pl.index_to_distance, 5.2)
    assert distance == 5.2


@pytest_benchmark
def test_distance_to_index(benchmark):
    pl = Polyline([Point(x) for x in range(10)])
    index = benchmark(pl.distance_to_index, 5.2)
    assert index == 5.2


def test_index_distance_reciprocal():
    for _ in range(10):
        pl = Polyline(np.random.random((10, 3)))
        index = np.random.random() * pl.get_total_distance()
        distance = pl.index_to_distance(index)
        assert pytest.approx(pl.distance_to_index(distance)) == index


@pytest_benchmark
def test_getitem_int(benchmark):
    pl = Polyline([Point(x) for x in range(10)])
    result = benchmark(pl.__getitem__, 5)
    assert result == Point(5, 0, 0)


@pytest_benchmark
def test_getitem_float(benchmark):
    pl = Polyline([Point(x) for x in range(10)])
    result = benchmark(pl.__getitem__, 3.4)
    assert pytest.approx(result.x) == 3.4


def test_getitem_float_zero_decimal():
    pl = Polyline([Point(x) for x in range(10)])
    assert pl[3.0].x == 3


def test_getitem_slice_integers():
    pl = Polyline([Point(x) for x in range(10)])
    assert np.all(pl[2:5].vertices == Polyline([Point(x) for x in range(2, 5)]).vertices)


def test_getitem_slice_floats():
    pl = Polyline([Point(x) for x in range(10)])
    assert np.all(pytest.approx(pl[2.2:5.9].vertices) == Polyline([Point(x) for x in [2.2, 3, 4, 5, 5.9]]).vertices)


def test_iter_polyline():
    pl = Polyline([Point(x) for x in range(10)])
    for i, point in enumerate(pl):
        assert pl[i].x == Point(i).x


@pytest_benchmark
def test_polyline_get_tangent_angles(benchmark):
    pl = Polyline([Point(x, x) for x in range(5)])
    tangents = benchmark(pl.get_tangent_angles)
    assert tangents.size == len(pl)
    assert np.all(tangents == pytest.approx(np.pi / 4))


@pytest_benchmark
def test_polyline_get_radius_of_curvature(benchmark):
    pl = Polyline(
        [Point(*spherical_to_cartesian(5, theta, np.pi / 2)) for theta in np.arange(-np.pi / 2, np.pi / 2, 0.1)]
    )
    radius_of_curvature = benchmark(pl.get_radius_of_curvature)
    assert radius_of_curvature.size == len(pl)
    assert np.all(radius_of_curvature == pytest.approx(5, rel=1e-2))


@pytest_benchmark
def test_polyline_resample(benchmark):
    pl = Polyline([Point(x, x) for x in range(5)])
    plr = benchmark(pl.resample, 0.5)
    assert pl.get_total_distance() == pytest.approx(plr.get_total_distance())
    assert np.all(plr.get_distances()[1:] - plr.get_distances()[:-1] == pytest.approx(0.5, rel=0.1))


def test_polyline_resample_return_indices():
    pl = Polyline([Point(x) for x in range(3)])
    plr, indices = pl.resample(0.5, return_indices=True)
    assert len(plr) == indices.size
    assert np.all(indices[1:] - indices[:-1] == pytest.approx(0.5))


@pytest_benchmark
def test_polyline_get_index_at_point(benchmark):
    pl = Polyline([Point(x) for x in range(3)])
    p = Point(1.3, 1)
    index = benchmark(pl.get_index_at_point, p)
    assert index == 1.3


@pytest_benchmark
def test_polyline_project_point(benchmark):
    pl = Polyline([Point(x) for x in range(3)])
    p = Point(1.3, 1)
    pp = benchmark(pl.project_point, p)
    assert pp == Point(1.3, 0.0, 0.0)



from openride.core.numba.transforms import transform_vertices, rotation_matrix

import numba
import numpy as np


@numba.njit(cache=True)
def bird_eye_view_box_vertices(
    position_x: float, position_y: float, size_x: float, size_y: float, yaw: float
) -> np.ndarray:
    """Get the 4 (x,y) coordinates of the corners of a bounding_box in bird eye view"""
    vertices = np.zeros((4, 3))
    vertices[:, 0] = size_x
    vertices[:, 1] = size_y
    vertices[2:, 0] *= -1
    vertices[1:3, 1] *= -1
    matrix4x4 = np.eye(4)
    matrix4x4[:3, :3] = rotation_matrix(0, 0, yaw)
    vertices = transform_vertices(vertices, matrix4x4)
    vertices[:, 0] += position_x
    vertices[:, 1] += position_y
    return vertices[:, :2]


@numba.njit(cache=True)
def box_vertices(
    position_x: float,
    position_y: float,
    position_z: float,
    size_x: float,
    size_y: float,
    size_z: float,
    roll: float,
    pitch: float,
    yaw: float,
) -> np.ndarray:
    """Get the 8 (x,y,z) coordinates of the corners of a bounding box in 3D"""
    vertices = np.zeros((8, 3))
    vertices[:, 0] = size_x
    vertices[:, 1] = size_y
    vertices[:, 2] = size_z
    vertices[:2, 0] *= -1
    vertices[6:, 0] *= -1
    vertices[4:, 1] *= -1
    vertices[::2, 2] *= -1
    matrix4x4 = np.eye(4)
    matrix4x4[:3, :3] = rotation_matrix(roll, pitch, yaw)
    vertices = transform_vertices(vertices, matrix4x4)
    vertices[:, 0] += position_x
    vertices[:, 1] += position_y
    vertices[:, 2] += position_z
    return vertices


@numba.njit(cache=True)
def vertices_inside_box(vertices: np.ndarray, box_vertices: np.ndarray) -> np.ndarray:
    """Returns indices of the vertices that are inside the box"""
    P = box_vertices
    u, v, w = P[1] - P[0], P[2] - P[0], P[6] - P[0]
    pu, pv, pw = np.dot(vertices, u), np.dot(vertices, v), np.dot(vertices, w)
    inside_box = np.where(
        (pu <= np.dot(P[1], u))
        & (pu >= np.dot(P[0], u))
        & (pv <= np.dot(P[2], v))
        & (pv >= np.dot(P[0], v))
        & (pw <= np.dot(P[6], w))
        & (pw >= np.dot(P[0], w))
    )[0]
    return inside_box



from openride import Point, BoundingBox, PointCloud, BoundingBoxCollection, Polyline

import cv2
import numpy as np

from typing import List, Tuple, Union



class ImageViewer:
    def __init__(
        self, title: str = "Openride Image Viewer", camera_matrix: np.ndarray = np.eye(3), distortion: List[float] = []
    ):
        self.title = title
        self.camera_matrix = camera_matrix
        self.distortion = distortion

        cv2.namedWindow(title, cv2.WINDOW_GUI_NORMAL)
        cv2.setMouseCallback(title, self.mouse_callback)

        self.display = None

    def set_window_size(self, size:Tuple[int, int]):
        cv2.resizeWindow(self.title, size[0], size[1])

    def update(self):
        if self.display is None:
            return
        cv2.imshow(self.title, self.display[..., ::-1])
        cv2.waitKey(1)

    def draw_image(self, image: np.ndarray):
        self.display = image / 256


    def draw_point(self, point: Point, color: tuple = (1, 1, 1), size: int = 3, alpha:float=1.0):

        if point.z < 0:
            return
        
        point2d = self.__project_points(np.array((point.x, point.y, point.z))).astype(int)

        if alpha >= 1.0:
            for p in point2d:
                cv2.circle(self.display, p, size, color, -1)

        elif alpha > 0.0:
            tmp = np.zeros_like(self.display)
            for p in point2d:
                cv2.circle(tmp, p, size, color, -1)
            mask = tmp.astype(bool)
            self.display[mask] = cv2.addWeighted(self.display, 1-alpha, tmp, alpha, 0)[mask]


    def labelize(self, label: str, pos: Point, color: tuple = (1,1,1), fontsize: float = 1.0, linewidth:int=1):
        point2d = self.__project_points(np.array((pos.x, pos.y, pos.z))).astype(int)
        if point2d.size == 0:
            return
        self.display = cv2.putText(self.display, label, point2d[0], cv2.FONT_HERSHEY_SIMPLEX, fontsize, color, linewidth)

    def labelize_2d(self, label: str, x:int, y:int, color: tuple = (1,1,1), fontsize: float = 1.0, linewidth:int=1):
        self.display = cv2.putText(self.display, label, [x, y], cv2.FONT_HERSHEY_SIMPLEX, fontsize, color, linewidth)


    def draw_rectangle(self, xmin:float, ymin:float, xmax:float, ymax:float, color:tuple=(1,1,1), linewidth:int=1, fill:bool=False, alpha:float=1.0):
        _linewidth = linewidth if not fill else -1 * linewidth
        if alpha >= 1.0:
            self.display = cv2.rectangle(self.display, (int(xmin), int(ymin)), (int(xmax), int(ymax)), color, _linewidth)
        elif alpha > 0.0:
            tmp = np.zeros_like(self.display)
            tmp = cv2.rectangle(tmp, (int(xmin), int(ymin)), (int(xmax), int(ymax)), color, _linewidth)
            mask = tmp.astype(bool)
            self.display[mask] = cv2.addWeighted(self.display, 1-alpha, tmp, alpha, 0)[mask]


    def draw_point_cloud(
        self,
        pcloud: PointCloud,
        color: tuple = (1, 1, 1),
        alpha: float = 1.0,
    ):
        mask = np.where(pcloud.get_point_cloud()[:, 2] > 0.0)[0]
        points2d = self.__project_points(pcloud.get_point_cloud()[mask]).astype(int)

        points2d = points2d[np.where(points2d[:,0] >= 0)]
        points2d = points2d[np.where(points2d[:,1] >= 0)]
        points2d = points2d[np.where(points2d[:,0] < self.display.shape[1]-1)]
        points2d = points2d[np.where(points2d[:,1] < self.display.shape[0]-1)]

        if alpha >= 1.0:
            self.display[points2d[:,1], points2d[:,0]] = color
        elif alpha > 0.0:
            tmp = np.zeros_like(self.display)
            tmp[points2d[:,1], points2d[:,0]] = color
            mask = tmp.astype(bool)
            self.display[mask] = cv2.addWeighted(self.display, 1-alpha, tmp, alpha, 0)[mask]


    def draw_bounding_box(
        self,
        bounding_box: Union[BoundingBox, BoundingBoxCollection],
        color: tuple = (1, 1, 1),
        linewidth: int = 2,
        alpha: float = 1.0,
    ):
        if isinstance(bounding_box, BoundingBoxCollection):
            [self.draw_bounding_box(box, color, linewidth, alpha) for box in bounding_box]
            return

        lines = np.array(self.__get_box_projection(bounding_box))

        if alpha >= 1.0:    
            cv2.polylines(self.display, lines, False, color, linewidth)
        elif alpha > 0.0:
            tmp = np.zeros_like(self.display)
            cv2.polylines(tmp, lines, False, color, linewidth)
            mask = tmp.astype(bool)
            self.display[mask] = cv2.addWeighted(self.display, 1-alpha, tmp, alpha, 0)[mask]


    def draw_polyline(
        self,
        line: Polyline,
        color: tuple = (1, 1, 1),
        linewidth: int = 2,
        alpha :float = 1.0,
    ):
                
        if len(line) < 2:
            return
        
        for i in range(len(line)-1):
            if np.sign(line[i].z) != np.sign(line[i+1].z):
                new_line = Polyline([line[i], line[i+1]]).resample(1)
                new_line.vertices = new_line.vertices[np.where(new_line.vertices[:, 2] > 0.0)[0]]
                self.draw_polyline(new_line, color, linewidth, alpha)

        mask = np.where(line.vertices[:, 2] > 0.0)[0]
        if mask.shape[0] < 2:
            return
        points2d = self.__project_points(line.vertices[mask]).astype(int)

        if alpha >= 1.0:
            cv2.polylines(self.display, [points2d.reshape((-1, 1, 2))], False, color, linewidth)
        elif alpha > 0.0:
            tmp = np.zeros_like(self.display)
            cv2.polylines(self.display, [points2d.reshape((-1, 1, 2))], False, color, linewidth)
            mask = tmp.astype(bool)
            self.display[mask] = cv2.addWeighted(self.display, 1-alpha, tmp, alpha, 0)[mask]


    def mouse_callback(self, event, x, y, *_):
        pass

    def close(self):
        cv2.destroyWindow(self.title)

    def __project_points(self, points: np.ndarray) -> np.ndarray:
        if points.shape[0] == 0:
            return np.empty((0, 2))
        R = T = np.zeros((3, 1))
        pts2d, _ = cv2.projectPoints(
            points.astype(float), R, T, self.camera_matrix, np.array(self.distortion)
        )
        return pts2d[:, 0, :]

    def __get_box_projection(self, box: BoundingBox) -> List[Tuple[Tuple[int, int], Tuple[int, int]]]:

        lines: List[Tuple[Tuple[int, int], Tuple[int, int]]] = []
        vertices = box.get_vertices()
        mask = vertices[:, 2] >= 0
        pts2d = self.__project_points(vertices)

        indices = [6, 7, 7, 1, 1, 0, 0, 6, 4, 5, 5, 3, 3, 2, 2, 4, 6, 4, 7, 5, 0, 2, 1, 3]
        for i in range(0, len(indices) - 1, 2):
            if not mask[indices[i]] or not mask[indices[i + 1]]:
                continue
            p1 = tuple(pts2d[indices[i]].astype(int))
            p2 = tuple(pts2d[indices[i + 1]].astype(int))
            lines.append((p1, p2))
        return lines
    

    def set_camera_matrix(self, matrix:np.ndarray):
        self.camera_matrix = matrix

    def set_fov(self, resolution:Tuple[int, int], fov_degrees:float):
        matrix = np.identity(3)
        matrix[0,2] = resolution[0]/2
        matrix[1,2] = resolution[1]/2
        matrix[0,0] = matrix[1,1] = resolution[0]/(2*np.tan(fov_degrees * np.pi / 360.0))
        self.set_camera_matrix(matrix)



from openride import BoundingBoxCollection, BoundingBox, Point, Rotation, Size, Transform
from openride.test import get_pytest_benchmark

import numpy as np
import pytest


pytest_benchmark = get_pytest_benchmark(group="core.collections.bounding_box_collection")


def test_init_empty():
    bc = BoundingBoxCollection()
    assert len(bc) == 0


def test_init():
    bc = BoundingBoxCollection([BoundingBox(), BoundingBox()])
    assert len(bc) == 2


def test_append():
    bc = BoundingBoxCollection([BoundingBox(), BoundingBox()])
    bc.append(BoundingBox())
    assert len(bc) == 3


def test_extend():
    bc = BoundingBoxCollection([BoundingBox(), BoundingBox()])
    bc2 = BoundingBoxCollection([BoundingBox(), BoundingBox()])
    bc.extend(bc2)
    assert len(bc) == 4


def test_pop():
    bc = BoundingBoxCollection([BoundingBox(), BoundingBox(Point(1)), BoundingBox()])
    box = bc.pop(1)
    assert box.position.x == 1
    assert len(bc) == 2


def test_repr():
    bc = BoundingBoxCollection([BoundingBox(), BoundingBox()])
    assert isinstance(str(bc), str)


@pytest_benchmark
def test_getitem(benchmark):
    bc = BoundingBoxCollection(
        [
            BoundingBox(),
            BoundingBox(Point(1), Rotation(0, 2), Size(1, 1, 3)),
            BoundingBox(),
        ]
    )
    box = benchmark(bc.__getitem__, 1)
    assert box.position.x == 1
    assert box.rotation.pitch == 2
    assert box.size.z == 3


@pytest_benchmark
def test_transform(benchmark):
    bc = BoundingBoxCollection([BoundingBox() for _ in range(10)])
    tf = Transform(Point(1), Rotation(0, 0, np.pi))
    transformed = benchmark(bc.transform, tf)
    assert pytest.approx(transformed[0].position.x) == 1
    assert pytest.approx(transformed[1].position.x) == 1
    assert pytest.approx(transformed[0].rotation.yaw) == np.pi
    assert transformed[0].size == bc[0].size
    assert bc[0].position.x == 0.0  # original is untouched


@pytest_benchmark
def test_get_distances(benchmark):
    bc = BoundingBoxCollection([BoundingBox(Point(x, x, x)) for x in range(10)])
    distances = benchmark(bc.get_distances)
    assert np.all(distances[:3] == np.array([0, 3**0.5, 12**0.5]))


def test_filter():
    bc = BoundingBoxCollection([BoundingBox(Point(x)) for x in range(3)])
    bcf = bc.filter([0, 2])
    assert bcf[0].position.x == 0 and bcf[1].position.x == 2


def test_filter_distance_min():
    bc = BoundingBoxCollection([BoundingBox(Point(x, x, x)) for x in range(3)])
    bcf = bc.filter_distance(min_distance=2.0)
    assert bcf.get_distances() == np.array([12**0.5])


@pytest_benchmark
def test_filter_distance_max(benchmark):
    bc = BoundingBoxCollection([BoundingBox(Point(x, x, x)) for x in range(30)])
    bcf = benchmark(bc.filter_distance, max_distance=10.0)
    assert np.max(bcf.get_distances()) <= 10.0



from openride.core.geometry import Geometry
from openride.core.point import Point
from openride.core.rotation import Rotation
from openride.core.size import Size
from openride.core.transform import Transform
from openride.core.numba import (
    bird_eye_view_box_vertices,
    box_vertices,
)

from dataclasses import dataclass, field
from shapely import geometry

import numpy as np


@dataclass
class BoundingBox(Geometry):
    """3D bounding box defined by a center position, rotation (euler angles) and size."""

    position: Point = field(default_factory=Point)
    rotation: Rotation = field(default_factory=Rotation)
    size: Size = field(default_factory=Size)

    @classmethod
    def from_raw(cls, 
                 x:float=0.0, y:float=0.0, z:float=0.0, 
                 roll:float=0.0, pitch:float=0.0, yaw:float=0.0, 
                 sx:float=1.0, sy:float=1.0, sz:float=1.0,
        ) -> 'BoundingBox':
        return cls(Point(x, y, z), Rotation(roll, pitch, yaw), Size(sx, sy, sz))

    def to_shapely(self) -> geometry.Polygon:
        return geometry.Polygon(self.get_bird_eye_view_vertices())

    def transform(self, transform: "Transform") -> "BoundingBox":
        return BoundingBox(
            self.position.transform(transform),
            self.rotation + transform.rotation,
            self.size,
        )

    def get_transform(self) -> Transform:
        return Transform(self.position, self.rotation)

    def get_bird_eye_view_vertices(self) -> np.ndarray:
        return bird_eye_view_box_vertices(
            self.position.x,
            self.position.y,
            self.size.x,
            self.size.y,
            self.rotation.yaw,
        )

    def get_vertices(self) -> np.ndarray:
        return box_vertices(
            self.position.x,
            self.position.y,
            self.position.z,
            self.size.x,
            self.size.y,
            self.size.z,
            self.rotation.roll,
            self.rotation.pitch,
            self.rotation.yaw,
        )



from openride import Point, Rotation, Transform
from openride.test import get_pytest_benchmark
from openride.test.random_core_generator import get_random

import numpy as np
import pytest


pytest_benchmark = get_pytest_benchmark(group="core.transform")


def test_transform_init():
    tf = Transform()
    assert tf.translation == Point()
    assert tf.rotation == Rotation()


@pytest_benchmark
def test_transform_get_matrix(benchmark):
    tf = Transform()
    matrix = benchmark(tf.get_matrix)
    assert np.all(matrix == np.eye(4))


def test_transform_matrix_translation():
    tf = Transform(translation=Point(1, 2, 3))
    matrix = tf.get_matrix()
    assert matrix[0, 3] == 1
    assert matrix[1, 3] == 2
    assert matrix[2, 3] == 3


@pytest_benchmark
def test_transform_get_inverse_matrix(benchmark):
    tf = Transform()
    matrix = benchmark(tf.get_inverse_matrix)
    assert np.all(matrix == np.eye(4))


def test_transform_inverse_matrix_translation():
    tf = Transform(translation=Point(1, 2, 3))
    matrix = tf.get_inverse_matrix()
    assert matrix[0, 3] == -1
    assert matrix[1, 3] == -2
    assert matrix[2, 3] == -3


def test_transform_matrix_inverse_reciprocal():
    tf1 = Transform(
        translation=Point(1, 2, 3),
        rotation=Rotation(*np.random.random(3)),
    )
    tf2 = Transform.from_matrix(tf1.get_matrix())
    assert np.all(pytest.approx(tf1.get_matrix()) == tf2.get_matrix())


@pytest_benchmark
def test_transform_inverse(benchmark):
    tf = Transform(
        translation=Point(1, 2, 3),
        rotation=Rotation(*np.random.random(3)),
    )
    tf_inf = benchmark(tf.inverse)
    p1 = Point(8, 3, 4)
    p2 = p1.transform(tf).transform(tf_inf)
    assert pytest.approx(p1.x) == p2.x
    assert pytest.approx(p1.y) == p2.y
    assert pytest.approx(p1.z) == p2.z


def test_matrix_mul_inverse_equals_identity():
    for _ in range(5):
        tf = get_random(Transform)
        mat = tf.get_matrix()
        inv = tf.get_inverse_matrix()
        res = mat @ inv
        assert res == pytest.approx(np.eye(4))


def test_init_from_6dof():
    tf1 = Transform.from_6dof()
    tf2 = Transform()
    assert tf1 == tf2



from openride.core.point import Point
from openride.core.rotation import Rotation
from openride.core.numba import transform_matrix, transform_inverse_matrix

from dataclasses import dataclass, field

import numpy as np


@dataclass
class Transform:

    translation: Point = field(default_factory=Point)
    rotation: Rotation = field(default_factory=Rotation)

    def get_matrix(self) -> np.ndarray:
        return transform_matrix(
            self.translation.x,
            self.translation.y,
            self.translation.z,
            self.rotation.roll,
            self.rotation.pitch,
            self.rotation.yaw,
        )

    def get_inverse_matrix(self) -> np.ndarray:
        return transform_inverse_matrix(
            self.translation.x,
            self.translation.y,
            self.translation.z,
            self.rotation.roll,
            self.rotation.pitch,
            self.rotation.yaw,
        )

    def inverse(self) -> "Transform":
        return Transform.from_matrix(self.get_inverse_matrix())

    @classmethod
    def from_matrix(cls, matrix: np.ndarray) -> "Transform":
        return cls(Point(*matrix[:3, 3]), Rotation.from_matrix(matrix))
    
    @classmethod
    def from_6dof(cls, 
                  x:float=0.0, y:float=0.0, z:float=0.0, 
                  roll:float=0.0, pitch:float=0.0, yaw:float=0.0, 
        ) -> 'Transform':
        return cls(Point(x, y, z), Rotation(roll, pitch, yaw))



from openride import Category, SubCategory, Rotation

from typing import List, Optional, Union

import glob
import os
import numpy as np

FILEPATH = os.path.dirname(os.path.abspath(__file__))


DEFAULT_MODELS = {
    Category.Vehicle: os.path.join(FILEPATH, "cars", "Aston_martin.stl"),
    Category.Pedestrian: os.path.join(FILEPATH, "pedestrians", "default_male.stl"),
    Category.Cyclist: os.path.join(FILEPATH, "cyclists", "bicycle2.stl"),
    Category.TrafficCone: os.path.join(FILEPATH, "traffic_cones", "cone.stl"),
    SubCategory.Car: os.path.join(FILEPATH, "cars", "Aston_martin.stl"),
    SubCategory.Van: os.path.join(FILEPATH, "vans", "volkswagen_transporter.stl"),
    SubCategory.Bus: os.path.join(FILEPATH, "buses", "ShortSchoolBus.stl"),
    SubCategory.Truck: os.path.join(FILEPATH, "trucks", "Truck2.stl"),
    SubCategory.Motorcycle: os.path.join(FILEPATH, "motorcycles", "kawaski.stl"),
}


DIRECTORIES = {
    Category.Vehicle: os.path.join(FILEPATH, "cars"),
    Category.Pedestrian: os.path.join(FILEPATH, "pedestrians"),
    Category.Cyclist: os.path.join(FILEPATH, "cyclists"),
    Category.TrafficCone: os.path.join(FILEPATH, "traffic_cones"),
    SubCategory.Car: os.path.join(FILEPATH, "cars"),
    SubCategory.Van: os.path.join(FILEPATH, "vans"),
    SubCategory.Bus: os.path.join(FILEPATH, "buses"),
    SubCategory.Truck: os.path.join(FILEPATH, "trucks"),
    SubCategory.Motorcycle: os.path.join(FILEPATH, "motorcycles"),
}


def get_model_files(category: Optional[Union[Category, SubCategory]]=None) -> List[str]:
    if category is None:
        return glob.glob(os.path.join(FILEPATH, "*", "*.stl"))
    directory = DIRECTORIES.get(category)
    if not directory:
        return []
    return glob.glob(f"{directory}/*")


MODEL_ROTATIONS = {
    "jeep_renegate.stl": Rotation(yaw=np.pi / 2),
    "Aston_martin.stl": Rotation(yaw=np.pi),
    "chevrolet_camaro_ss.stl": Rotation(yaw=np.pi / 2),
    "chevrolet_corvette_zr1.stl": Rotation(yaw=np.pi / 2),
    "dodge_challenger.stl": Rotation(yaw=np.pi / 2),
    "mitsubishi_lancer_evolution.stl": Rotation(yaw=np.pi / 2),
    "nissan_skyline_gtr.stl": Rotation(yaw=np.pi / 2),
    "shelby_mustang_gt350r.stl": Rotation(yaw=np.pi / 2),
    "tesla_model_3.stl": Rotation(yaw=np.pi / 2),
    "volkswagen_golf_gti.stl": Rotation(yaw=np.pi / 2),
    "chopper_reduziert.stl": Rotation(roll=np.pi / 2),
    "ducati.stl": Rotation(yaw=np.pi / 2),
    "kawaski.stl": Rotation(yaw=np.pi / 2),
    "volkswagen_transporter.stl": Rotation(yaw=np.pi / 2),
    "vw_transporter2.stl": Rotation(yaw=np.pi / 2),
    "Fire_Truck.stl": Rotation(),
    "trashtruck.stl": Rotation(),
    "Truck2.stl": Rotation(),
    "ShortSchoolBus.stl": Rotation(),
    "default_male.stl": Rotation(pitch = np.pi/4, yaw=np.pi / 2),
    "sophie_walk.stl": Rotation(pitch=np.pi/2, yaw=np.pi/2),
    "tyler_walk.stl": Rotation(pitch=np.pi / 2, yaw=np.pi / 2),
    "woman_stand.stl": Rotation(yaw=np.pi / 2),
    "default_female.stl": Rotation(yaw=np.pi / 2),
    "vahlen_stand.stl": Rotation(yaw=np.pi / 2),
    "bicycle.stl": Rotation(),
    "bicycle2.stl": Rotation(),
}



from openride.core.numba.box_utils import bird_eye_view_box_vertices, box_vertices, vertices_inside_box
from openride.core.numba.coordinate_systems import cartesian_to_spherical, spherical_to_cartesian
from openride.core.numba.distances_3d import (
    distance_between_points_3d,
    distance_between_box_and_point_3d,
    distance_between_line_and_point_3d,
    distance_between_lines_3d,
    distance_between_line_and_box_3d,
    distance_between_boxes_3d,
)
from openride.core.numba.distances_bev import (
    bev_distance_between_points,
    bev_distance_between_box_and_point,
    bev_project_point_on_line,
    bev_distance_between_line_and_point,
    bev_distance_between_polygon_and_point,
    bev_distance_between_line_and_polygon,
    bev_distance_between_line_and_box,
    bev_distance_between_polygon_and_box,
    bev_distance_between_polygons,
    bev_distance_between_boxes,
    bev_distance_between_lines,
)
from openride.core.numba.interpolation import (
    linear_interpolation,
    linear_interpolation_angle,
    interpolate_line_index_at_xy,
)
from openride.core.numba.polygon_utils import polygon_contains_point, polygon_contains_all_vertices
from openride.core.numba.polyline_utils import (
    vertices_to_distances,
    index_to_distance,
    distance_to_index,
    left_or_right_from_line,
    tangent_angle,
    tangent_angles,
    radius_of_curvature,
    resample_linear,
    project_point_on_line,
)
from openride.core.numba.transforms import (
    rotation_matrix,
    transform_matrix,
    transform_inverse_matrix,
    euler_angles,
    transform_point,
    transform_vertices,
)



from openride import Viewer
from typing import Tuple

import numpy as np
import os
import cv2


def image_to_ascii_art(img: np.ndarray, resolution: Tuple[int, int]) -> str:
    """Convert an Image to ASCII Art"""

    pixels = cv2.resize(img, resolution)
    pixels = np.mean(pixels, axis=-1).astype(int)

    chars = ["*", "S", "#", "&", "@", "$", "%", "*", "!", ":", "."]
    new_pixels = [chars[pixel // 25] for pixel in np.ravel(pixels)]
    new_pixels = "".join(new_pixels)

    new_pixels_count = len(new_pixels)
    ascii_image = [new_pixels[index : index + resolution[0]] for index in range(0, new_pixels_count, resolution[0])]
    ascii_image = "\n".join(ascii_image)

    return ascii_image


class AsciiViewer(Viewer):
    def __init__(self, resolution: Tuple[int, int] = (80, 40)):
        super().__init__(
            resolution=resolution, background=(0, 0, 0), mouse_camera_interactions=False, render_offscreen=True
        )

    def update(self):
        rendered_image = super().update(return_image=True)
        ascii_image = image_to_ascii_art(rendered_image, self._resolution)

        clear_console = "clear" if os.name == "posix" else "CLS"
        os.system(clear_console)
        print(ascii_image)



from dataclasses import dataclass


@dataclass
class Size:

    x: float = 1.0
    y: float = 1.0
    z: float = 1.0

    def __post_init__(self):
        if self.x < 0 or self.y < 0 or self.z < 0:
            raise ValueError("Size can't be negative")



from openride import Point, Polygon, Transform
from openride.test import get_pytest_benchmark
from openride.test.random_core_generator import get_random

from shapely import geometry

import numpy as np
import pytest


pytest_benchmark = get_pytest_benchmark(group="core.polygon")


def test_init_list_points():
    p = Polygon([Point(x) for x in range(10)] + [Point(0)])
    assert p.vertices.shape == (11, 3)


def test_post_init_connect_ends():
    p = get_random(Polygon)
    assert np.all(p.vertices[0] == p.vertices[-1])


def test_polygon_to_shapely():
    assert isinstance(get_random(Polygon).to_shapely(), geometry.Polygon)


@pytest_benchmark
def test_polygon_transform_identity(benchmark):
    polygon = get_random(Polygon)
    tf = Transform()
    polygon_tf = benchmark(polygon.transform, tf)
    assert np.all(polygon.vertices == polygon_tf.vertices)


def test_polygon_from_shapely():
    p1 = get_random(Polygon)
    p2 = Polygon.from_shapely(p1.to_shapely())
    assert np.all(pytest.approx(p1.vertices) == p2.vertices)


@pytest_benchmark
def test_polygon_area(benchmark):
    poly = Polygon([Point(1, 1), Point(-1, 1), Point(-1, -1), Point(1, -1)])
    assert benchmark(poly.get_area) == 4.0



import numba
import numpy as np


@numba.njit(cache=True)
def polygon_contains_point(vertices: np.ndarray, x: float, y: float) -> bool:
    n = len(vertices)
    inside = False
    p2x = 0.0
    p2y = 0.0
    xints = 0.0
    p1x, p1y = vertices[0, :2]
    for i in range(n + 1):
        p2x, p2y = vertices[i % n, :2]
        if y > min(p1y, p2y):
            if y <= max(p1y, p2y):
                if x <= max(p1x, p2x):
                    if p1y != p2y:
                        xints = (y - p1y) * (p2x - p1x) / (p2y - p1y) + p1x
                    if p1x == p2x or x <= xints:
                        inside = not inside
        p1x, p1y = p2x, p2y

    return inside


@numba.njit(cache=True)
def polygon_contains_all_vertices(poly_vertices: np.ndarray, vertices: np.ndarray) -> bool:
    for v in vertices[:, :2]:
        if not polygon_contains_point(poly_vertices, v[0], v[1]):
            return False
    return True



from openride import Size
import pytest


def test_init():
    s = Size(1, 2, 3)
    assert s.x == 1
    assert s.y == 2
    assert s.z == 3


def test_post_init_fail():
    with pytest.raises(ValueError):
        Size(-1)



from openride import Point, BoundingBox, Polygon, Polyline, Size, contains
from openride.test import get_pytest_benchmark


pytest_benchmark = get_pytest_benchmark(group="operations.contains_3D")


@pytest_benchmark
def test_contains_box_point_true(benchmark):
    box = BoundingBox()
    point = Point()
    assert benchmark(contains, box, point)


def test_contains_box_point_false():
    assert not contains(BoundingBox(), Point(0, 0, 5))


def test_contains_box_point_on_edge():
    assert contains(BoundingBox(), Point(1))


@pytest_benchmark
def test_contains_box_box_true(benchmark):
    box1 = BoundingBox(size=Size(1, 1, 1))
    box2 = BoundingBox(size=Size(0.1, 0.1, 0.1))
    assert benchmark(contains, box1, box2)


def test_contains_box_box_false():
    assert not contains(BoundingBox(size=Size(1, 1, 1)), BoundingBox(size=Size(2, 2, 2)))


def test_contains_box_box_partial_false():
    assert not contains(BoundingBox(), BoundingBox(Point(1, 1, 1)))


@pytest_benchmark
def test_contains_box_polyline_true(benchmark):
    box = BoundingBox()
    polyline = Polyline([Point(), Point(0.5)])
    assert benchmark(contains, box, polyline)


def test_contains_box_polyline_false():
    assert not contains(BoundingBox(), Polyline([Point(), Point(1.5)]))


@pytest_benchmark
def test_contains_box_polygon_true(benchmark):
    box = BoundingBox()
    polygon = Polygon([Point(), Point(0.5), Point(0.5, 0.5)])
    assert benchmark(contains, box, polygon)


def test_contains_box_polygon_false():
    assert not contains(BoundingBox(), Polygon([Point(), Point(1.5), Point(1.5, 1.5)]))


@pytest_benchmark
def test_contains_polyline_point_true(benchmark):
    polyline = Polyline([Point(x, x, x) for x in range(2)])
    point = Point(0.5, 0.5, 0.5)
    assert benchmark(contains, polyline, point)


def test_contains_polyline_point_false():
    pl = Polyline([Point(x, x, x) for x in range(2)])
    assert not contains(pl, Point(0.5, 0.5, 0.0))


@pytest_benchmark
def test_contains_polyline_polyline_true(benchmark):
    pl1 = Polyline([Point(x, x, x) for x in range(5)])
    pl2 = Polyline([Point(x, x, x) for x in range(1, 4)])
    assert benchmark(contains, pl1, pl2)


def test_contains_polyline_polyline_false():
    pl1 = Polyline([Point(x, x, x) for x in range(5)])
    pl2 = Polyline([Point(x, x, x) for x in range(1, 6)])
    assert not contains(pl1, pl2)



from openride import Point, BoundingBox, Polyline, Rotation, Size, distance
from openride.test import get_pytest_benchmark

import numpy as np
import pytest


pytest_benchmark = get_pytest_benchmark(group="operations.distance_3D")


@pytest_benchmark
def test_distance_point_point(benchmark):
    p1 = Point(1, 1, 1)
    p2 = Point(2, 2, 2)
    assert benchmark(distance, p1, p2) == 3**0.5


@pytest_benchmark
def test_distance_point_box(benchmark):
    b = BoundingBox()
    p = Point(2.0, 2.0, 2.0)
    assert benchmark(distance, p, b) == 3**0.5
    assert distance(b, p) == 3**0.5


def test_distance_point_box_overlap():
    b = BoundingBox()
    p = Point(0.5, 0.5, 0.5)
    assert distance(p, b) == 0.0
    assert distance(b, p) == 0.0


def test_distance_point_box_edge():
    b = BoundingBox()
    p = Point(2.0, 0.0, 2.0)
    assert distance(p, b) == 2**0.5
    assert distance(b, p) == 2**0.5


def test_distance_point_box_face():
    b = BoundingBox()
    p = Point(0.75, 0.5, 2)
    assert distance(p, b) == 1.0
    assert distance(b, p) == 1.0


@pytest_benchmark
def test_distance_point_line(benchmark):
    l = Polyline([Point(0, 0, 0), Point(1, 0, 1)])
    p = Point(0.5, 0.5, 0.5)
    assert benchmark(distance, l, p) == pytest.approx(0.5)
    assert distance(p, l) == pytest.approx(0.5)


def test_distance_point_line_overlap():
    l = Polyline([Point(0, 0, 0), Point(1, 1, 1)])
    p = Point(0.75, 0.75, 0.75)
    assert distance(l, p) == 0.0
    assert distance(p, l) == 0.0


def test_distance_point_line_far():
    l = Polyline([Point(0, 0, 0), Point(1, 1, 1)])
    p = Point(2, 2, 2)
    assert pytest.approx(distance(l, p)) == 3**0.5
    assert pytest.approx(distance(p, l)) == 3**0.5


@pytest_benchmark
def test_distance_box_box(benchmark):
    b1 = BoundingBox(Point(0, 0, 0), Rotation(0, 0, 0), Size(1, 1, 1))
    b2 = BoundingBox(Point(3, 3, 3), Rotation(0, 0, 0), Size(1, 1, 1))
    assert benchmark(distance, b1, b2) == 3**0.5
    assert distance(b2, b1) == 3**0.5


def test_distance_box_box_overlap():
    b1 = BoundingBox()
    b2 = BoundingBox(Point(0, 0, 2), Rotation(np.pi / 4, np.pi / 4, 0))
    assert distance(b1, b2) == 0.0
    assert distance(b2, b1) == 0.0


def test_distance_box_box_corner_to_face():
    b1 = BoundingBox()
    b2 = BoundingBox(Point(0, 0, 3), Rotation(np.pi / 4, np.pi / 4, 0))
    assert pytest.approx(distance(b1, b2)) == (2 - 2**0.5) / 2
    assert pytest.approx(distance(b2, b1)) == (2 - 2**0.5) / 2


def test_distance_box_box_edge_to_face():
    b1 = BoundingBox()
    b2 = BoundingBox(Point(0, 0, 5), Rotation(np.pi / 4, 0, 0), Size(2, 2, 2))
    assert pytest.approx(distance(b1, b2)) == 4 - 2 * 2**0.5
    assert pytest.approx(distance(b2, b1)) == 4 - 2 * 2**0.5


@pytest_benchmark
def test_distance_box_line(benchmark):
    l = Polyline([Point(-1, 2, 2), Point(0, 2, 2), Point(1, 2, 2)])
    b = BoundingBox()
    assert benchmark(distance, l, b) == pytest.approx(2**0.5)
    assert distance(b, l) == pytest.approx(2**0.5)


def test_distance_box_line_overlap():
    l = Polyline([Point(0.5, 0.5, 0.5), Point(0, 2, 2), Point(1, 2, 2)])
    b = BoundingBox()
    assert distance(l, b) == 0.0
    assert distance(b, l) == 0.0


@pytest_benchmark
def test_distance_line_line(benchmark):
    l1 = Polyline([Point(0), Point(1), Point(2)])
    l2 = Polyline([Point(0.5, 1, 1), Point(1.5, 1, 1), Point(2.5, 1, 1)])
    assert benchmark(distance, l1, l2) == 2**0.5


def test_distance_line_line_far():
    l1 = Polyline([Point(0), Point(1), Point(2)])
    l2 = Polyline([Point(3, 1, 1), Point(4, 1, 1), Point(5, 1, 1)])
    assert distance(l1, l2) == 3**0.5


def test_distance_line_line_overlap():
    l1 = Polyline([Point(-1, -1, -1), Point(0, 0, 0), Point(1, 1, 1)])
    l2 = Polyline([Point(-1.5, 0, 0), Point(-0.5, 0, 0), Point(0.5, 0, 0)])
    assert distance(l1, l2) == 0.0



from openride import Point, Rotation, Transform
from openride.test import get_pytest_benchmark

from shapely import geometry

import numpy as np
import pytest


pytest_benchmark = get_pytest_benchmark(group="core.point")


def test_point_init():
    p = Point(1, 2, 3)
    assert p.x == 1 and p.y == 2 and p.z == 3


def test_point_to_shapely():
    p = Point(1, 2, 3)
    ps = p.to_shapely()
    assert isinstance(ps, geometry.Point)


@pytest_benchmark
def test_transform_point(benchmark):
    p = Point(0, 0, 0)
    tf = Transform()
    p2 = benchmark(p.transform, tf)
    assert p == p2


def test_translate_point():
    p = Point(1, 2, 3)
    tf = Transform(Point(4, 5, 6))
    p2 = p.transform(tf)
    assert p2 == Point(5, 7, 9)


def test_rotate_x_point():
    p = Point(1, 1, 1)
    tf = Transform(rotation=Rotation(roll=np.pi / 2))
    p2 = p.transform(tf)
    assert pytest.approx(p2.x) == 1
    assert pytest.approx(p2.y) == -1
    assert pytest.approx(p2.z) == 1


def test_rotate_y_point():
    p = Point(1, 1, 1)
    tf = Transform(rotation=Rotation(pitch=np.pi / 2))
    p2 = p.transform(tf)
    assert pytest.approx(p2.x) == 1
    assert pytest.approx(p2.y) == 1
    assert pytest.approx(p2.z) == -1


def test_rotate_z_point():
    p = Point(1, 1, 1)
    tf = Transform(rotation=Rotation(yaw=np.pi / 2))
    p2 = p.transform(tf)
    assert pytest.approx(p2.x) == -1
    assert pytest.approx(p2.y) == 1
    assert pytest.approx(p2.z) == 1


@pytest_benchmark
def test_add_points(benchmark):
    p1 = Point(1, 2, 3)
    p2 = Point(4, 5, 6)
    assert benchmark(p1.__add__, p2) == Point(5, 7, 9)


def test_substract_points():
    p1 = Point(1, 2, 3)
    p2 = Point(4, 5, 6)
    assert p1 - p2 == Point(-3, -3, -3)


def test_multiply_point():
    p = Point(1, 2, 3)
    assert p * 2 == Point(2, 4, 6)


def test_reverse_multiply_point():
    p = Point(1, 2, 3)
    assert 2 * p == Point(2, 4, 6)


def test_repr_point():
    assert isinstance(str(Point()), str)



import openride as ride


def main():

    viewer = ride.Viewer()

    viewer.camera.set_position_xyz(-10, 0, 6)
    viewer.camera.set_focus_xyz(4, 0, 0)

    viewer.hud.toggle_fps_count()
    viewer.hud.toggle_cursor_position()

    viewer.draw_grid(delta=5, extent=50)
    viewer.draw_grid(delta=1, alpha=0.3, extent=50)

    point = ride.Point()

    bounding_box = ride.BoundingBox(
        position=ride.Point(5, 4, 1),
        rotation=ride.Rotation(0, 0, 1),
        size=ride.Size(2, 1, 1),
    )

    polyline = ride.Polyline([ride.Point(x + 2, -3) for x in range(10)])

    polygon = ride.Polygon([ride.Point(7, 1), ride.Point(2, 1), ride.Point(3, -1), ride.Point(7, -1, 0.3)])

    while True:
        viewer.draw_point(point, persist=False)
        viewer.labelize("allo", point)
        viewer.draw_bounding_box(bounding_box, color=(0, 1, 0), alpha=0.35)
        viewer.draw_model(ride.Category.Vehicle, bounding_box, color=(0,0,1))
        viewer.draw_polyline(polyline, color=(1, 0, 1), linewidth=3, points_size=10)
        viewer.draw_polygon(polygon, color=(1, 0, 0), alpha=0.7)
        viewer.update()


if __name__ == "__main__":
    main()



from openride.core.numba.box_utils import vertices_inside_box

import numpy as np
import numba

EPSILON = np.finfo(float).eps


@numba.njit(cache=True)
def distance_between_points_3d(x1: float, y1: float, z1: float, x2: float, y2: float, z2: float) -> float:
    """Minimum distance between two points in 3D"""
    return ((x1 - x2) ** 2 + (y1 - y2) ** 2 + (z1 - z2) ** 2) ** 0.5


BOX_TRIANGLE_INDICES = np.array(
    [
        [0, 1, 2],
        [1, 2, 3],
        [0, 2, 4],
        [0, 4, 6],
        [2, 3, 4],
        [3, 4, 5],
        [0, 1, 6],
        [1, 6, 7],
        [4, 5, 6],
        [5, 6, 7],
        [1, 3, 5],
        [1, 5, 7],
    ]
)


@numba.njit(cache=True)
def distance_between_box_and_point_3d(
    box_vertices: np.ndarray, box_x: float, box_y: float, box_z: float, px: float, py: float, pz: float
) -> float:
    """Minimum distance between a 3D bounding box and a point"""

    point = np.array((px, py, pz))
    if len(vertices_inside_box(np.reshape(point, (1, 3)), box_vertices)) > 0:
        return 0.0

    box_center = np.array([box_x, box_y, box_z])

    for triangle in BOX_TRIANGLE_INDICES:
        _triangle = np.vstack(
            (
                box_vertices[triangle[0]],
                box_vertices[triangle[1]],
                box_vertices[triangle[2]],
            )
        )
        if ray_triangle_intersect(box_center, point, _triangle):
            return distance_triangle_point_3d(_triangle, point)

    return distance_triangle_point_3d(_triangle, point)


@numba.njit(cache=True)
def distance_between_line_and_point_3d(vertices: np.ndarray, x: float, y: float, z: float) -> float:
    """Minimum distance between a polyline and a point in 3D"""
    distances = np.zeros(vertices.shape[0] - 1)
    p = np.array((x, y, z))
    for i in range(vertices.shape[0] - 1):
        distances[i] = point_segment_distance_3d(p, vertices[i], vertices[i + 1])
    return np.min(distances)


@numba.njit(cache=True)
def distance_between_lines_3d(vertices1: np.ndarray, vertices2: np.ndarray) -> float:
    """Minimum distance between two polylines in 3D"""
    distances = np.zeros((vertices1.shape[0] - 1, vertices2.shape[0] - 1))
    for i in range(vertices1.shape[0] - 1):
        for j in range(vertices2.shape[0] - 1):
            distances[i, j] = segments_distance_3d(vertices1[i], vertices1[i + 1], vertices2[j], vertices2[j + 1])
    return np.min(distances)


@numba.njit(cache=True)
def distance_between_line_and_box_3d(line_vertices: np.ndarray, box_vertices: np.ndarray) -> float:
    """Minimum distance between a polyline and a bounding box in 3D"""
    if len(vertices_inside_box(line_vertices, box_vertices)) > 0:
        return 0.0
    distances = np.zeros((12, line_vertices.shape[0] - 1))
    for i, face in enumerate(BOX_TRIANGLE_INDICES):
        triangle = np.vstack((box_vertices[face[0]], box_vertices[face[1]], box_vertices[face[2]]))
        for j in range(line_vertices.shape[0] - 1):
            distances[i, j] = segment_triangle_distance(line_vertices[j], line_vertices[j + 1], triangle)
            if distances[i, j] == 0.0:
                return 0.0
    return np.min(distances)


BOX_SEGMENT_INDICES = (6, 7, 7, 1, 1, 0, 0, 6, 4, 5, 5, 3, 3, 2, 2, 4, 6, 4, 7, 5, 0, 2, 1, 3)


@numba.njit(cache=True)
def distance_between_boxes_3d(vertices1: np.ndarray, vertices2: np.ndarray) -> float:
    """Minimum distance between two 3D bounding boxes"""
    if len(vertices_inside_box(vertices1, vertices2)) > 0:
        return 0.0
    if len(vertices_inside_box(vertices2, vertices1)) > 0:
        return 0.0
    distances = []
    for face in BOX_TRIANGLE_INDICES:
        triangle = np.vstack((vertices1[face[0]], vertices1[face[1]], vertices1[face[2]]))
        for p in vertices2:
            distances.append(distance_triangle_point_3d(triangle, p))
    for face in BOX_TRIANGLE_INDICES:
        triangle = np.vstack((vertices2[face[0]], vertices2[face[1]], vertices2[face[2]]))
        for p in vertices1:
            distances.append(distance_triangle_point_3d(triangle, p))

    for i in range(0, len(BOX_SEGMENT_INDICES) - 1, 2):
        v1 = vertices1[BOX_SEGMENT_INDICES[i]]
        v2 = vertices1[BOX_SEGMENT_INDICES[i + 1]]
        for j in range(0, len(BOX_SEGMENT_INDICES) - 1, 2):
            v3 = vertices2[BOX_SEGMENT_INDICES[j]]
            v4 = vertices2[BOX_SEGMENT_INDICES[j + 1]]
            distances.append(segments_distance_3d(v1, v2, v3, v4))

    return min(distances)


@numba.njit(cache=True)
def point_segment_distance_3d(p: np.ndarray, a: np.ndarray, b: np.ndarray) -> float:
    """Minimum distance between a line segment (finite length) and a point in 3D"""

    # normalized tangent vector
    diff = b - a
    norm = (diff[0] ** 2 + diff[1] ** 2 + diff[2] ** 2) ** 0.5
    d = np.divide(diff, norm)

    # signed parallel distance components
    s = np.dot(a - p, d)
    t = np.dot(p - b, d)

    # clamped parallel distance
    h = max([s, t, 0.0])

    # perpendicular distance component
    c = np.cross(p - a, d)

    return np.hypot(h, np.linalg.norm(c))


@numba.njit(cache=True)
def clip(a: float, min: float, max: float) -> float:
    if a < min:
        return min
    elif a > max:
        return max
    return a


@numba.njit(cache=True)
def segments_distance_3d(p11: np.ndarray, p12: np.ndarray, p21: np.ndarray, p22: np.ndarray) -> float:
    """Minimum distance between two line segments (finite lengths)"""
    r = p21 - p11
    u = p12 - p11
    v = p22 - p21

    ru = np.dot(r, u)
    rv = np.dot(r, v)
    uu = np.dot(u, u)
    uv = np.dot(u, v)
    vv = np.dot(v, v)

    det = uu * vv - uv * uv

    if det < EPSILON * uu * vv:
        s = ru / uu
        t = 0
    else:
        s = clip((ru * vv - rv * uv) / det, 0, 1)
        t = clip((ru * uv - rv * uu) / det, 0, 1)

    S = clip((t * uv + ru) / uu, 0, 1)
    T = clip((s * uv - rv) / vv, 0, 1)

    A = p11 + S * u
    B = p21 + T * v
    C = B - A
    return (C[0] ** 2 + C[1] ** 2 + C[2] ** 2) ** 0.5


### WIP ###
# def min_distance_triangles_point_3d(triangles: np.ndarray, points: np.ndarray) -> float:
#     """https://stackoverflow.com/questions/32342620/closest-point-projection-of-a-3d-point-to-3d-triangles-with-numpy-scipy"""

#     # Unpack triangle points
#     p0,p1,p2 = np.asarray(triangles).swapaxes(0,1)

#     # Calculate triangle edges
#     e0 = p1-p0
#     e1 = p2-p0
#     a = np.einsum('...i,...i', e0 , e0)
#     b = np.einsum('...i,...i', e0 , e1)
#     c = np.einsum('...i,...i', e1 , e1)

#     # Calculate determinant and denominator
#     det = a*c - b*b
#     invDet = 1. / (det + EPSILON)
#     denom = a-2*b+c

#     # Project to the edges
#     p  = p0-points[:,np.newaxis]
#     d = np.einsum('...i,...i', e0 , p)
#     e = np.einsum('...i,...i', e1 , p)
#     u = b*e - c*d
#     v = b*d - a*e

#     # Calculate numerators
#     bd = b+d
#     ce = c+e
#     numer0 = (ce - bd) / denom
#     numer1 = (c+e-b-d) / denom
#     da = -d/a
#     ec = -e/c


#     # Vectorize test conditions
#     m0 = u + v < det
#     m1 = u < 0
#     m2 = v < 0
#     m3 = d < 0
#     m4 = (a+d > b+e)
#     m5 = ce > bd

#     t0 =  m0 &  m1 &  m2 &  m3
#     t1 =  m0 &  m1 &  m2 & ~m3
#     t2 =  m0 &  m1 & ~m2
#     t3 =  m0 & ~m1 &  m2
#     t4 =  m0 & ~m1 & ~m2
#     t5 = ~m0 &  m1 &  m5
#     t6 = ~m0 &  m1 & ~m5
#     t7 = ~m0 &  m2 &  m4
#     t8 = ~m0 &  m2 & ~m4
#     t9 = ~m0 & ~m1 & ~m2

#     u = np.where(t0, np.clip(da, 0, 1), u)
#     v = np.where(t0, 0, v)
#     u = np.where(t1, 0, u)
#     v = np.where(t1, 0, v)
#     u = np.where(t2, 0, u)
#     v = np.where(t2, np.clip(ec, 0, 1), v)
#     u = np.where(t3, np.clip(da, 0, 1), u)
#     v = np.where(t3, 0, v)
#     u *= np.where(t4, invDet, 1)
#     v *= np.where(t4, invDet, 1)
#     u = np.where(t5, np.clip(numer0, 0, 1), u)
#     v = np.where(t5, 1 - u, v)
#     u = np.where(t6, 0, u)
#     v = np.where(t6, 1, v)
#     u = np.where(t7, np.clip(numer1, 0, 1), u)
#     v = np.where(t7, 1-u, v)
#     u = np.where(t8, 1, u)
#     v = np.where(t8, 0, v)
#     u = np.where(t9, np.clip(numer1, 0, 1), u)
#     v = np.where(t9, 1-u, v)

#     closest_points = (p0.T +  u[:, np.newaxis] * e0.T + v[:, np.newaxis] * e1.T).swapaxes(2,1)
#     distances_sq = (closest_points[:,:,0] - points[:,None,0])**2 + (closest_points[:,:,1] - points[:,None,1])**2 + (closest_points[:,:,2] - points[:,None,2])**2
#     return np.sqrt(np.ravel(distances_sq)[np.argmin(distances_sq)])


@numba.njit(cache=True)
def distance_triangle_point_3d(triangle: np.ndarray, p: np.ndarray) -> float:
    """https://gist.github.com/joshuashaffer/99d58e4ccbd37ca5d96e"""

    B = triangle[0, :]
    E0 = triangle[1, :] - B
    E1 = triangle[2, :] - B
    D = B - p
    a = np.dot(E0, E0)
    b = np.dot(E0, E1)
    c = np.dot(E1, E1)
    d = np.dot(E0, D)
    e = np.dot(E1, D)
    f = np.dot(D, D)

    det = a * c - b * b
    s = b * e - c * d
    t = b * d - a * e

    # Terible tree of conditionals to determine in which region of the diagram
    # shown above the projection of the point into the triangle-plane lies.
    if (s + t) <= det:
        if s < 0.0:
            if t < 0.0:
                # region4
                if d < 0:
                    t = 0.0
                    if -d >= a:
                        s = 1.0
                        sqrdistance = a + 2.0 * d + f
                    else:
                        s = -d / a
                        sqrdistance = d * s + f
                else:
                    s = 0.0
                    if e >= 0.0:
                        t = 0.0
                        sqrdistance = f
                    else:
                        if -e >= c:
                            t = 1.0
                            sqrdistance = c + 2.0 * e + f
                        else:
                            t = -e / c
                            sqrdistance = e * t + f

                            # of region 4
            else:
                # region 3
                s = 0
                if e >= 0:
                    t = 0
                    sqrdistance = f
                else:
                    if -e >= c:
                        t = 1
                        sqrdistance = c + 2.0 * e + f
                    else:
                        t = -e / c
                        sqrdistance = e * t + f
                        # of region 3
        else:
            if t < 0:
                # region 5
                t = 0
                if d >= 0:
                    s = 0
                    sqrdistance = f
                else:
                    if -d >= a:
                        s = 1
                        sqrdistance = a + 2.0 * d + f
                        # GF 20101013 fixed typo d*s ->2*d
                    else:
                        s = -d / a
                        sqrdistance = d * s + f
            else:
                # region 0
                invDet = 1.0 / det
                s = s * invDet
                t = t * invDet
                sqrdistance = s * (a * s + b * t + 2.0 * d) + t * (b * s + c * t + 2.0 * e) + f
    else:
        if s < 0.0:
            # region 2
            tmp0 = b + d
            tmp1 = c + e
            if tmp1 > tmp0:  # minimum on edge s+t=1
                numer = tmp1 - tmp0
                denom = a - 2.0 * b + c
                if numer >= denom:
                    s = 1.0
                    t = 0.0
                    sqrdistance = a + 2.0 * d + f
                    # GF 20101014 fixed typo 2*b -> 2*d
                else:
                    s = numer / denom
                    t = 1 - s
                    sqrdistance = s * (a * s + b * t + 2 * d) + t * (b * s + c * t + 2 * e) + f

            else:  # minimum on edge s=0
                s = 0.0
                if tmp1 <= 0.0:
                    t = 1
                    sqrdistance = c + 2.0 * e + f
                else:
                    if e >= 0.0:
                        t = 0.0
                        sqrdistance = f
                    else:
                        t = -e / c
                        sqrdistance = e * t + f
                        # of region 2
        else:
            if t < 0.0:
                # region6
                tmp0 = b + e
                tmp1 = a + d
                if tmp1 > tmp0:
                    numer = tmp1 - tmp0
                    denom = a - 2.0 * b + c
                    if numer >= denom:
                        t = 1.0
                        s = 0
                        sqrdistance = c + 2.0 * e + f
                    else:
                        t = numer / denom
                        s = 1 - t
                        sqrdistance = s * (a * s + b * t + 2.0 * d) + t * (b * s + c * t + 2.0 * e) + f

                else:
                    t = 0.0
                    if tmp1 <= 0.0:
                        s = 1
                        sqrdistance = a + 2.0 * d + f
                    else:
                        if d >= 0.0:
                            s = 0.0
                            sqrdistance = f
                        else:
                            s = -d / a
                            sqrdistance = d * s + f
            else:
                # region 1
                numer = c + e - b - d
                if numer <= 0:
                    s = 0.0
                    t = 1.0
                    sqrdistance = c + 2.0 * e + f
                else:
                    denom = a - 2.0 * b + c
                    if numer >= denom:
                        s = 1.0
                        t = 0.0
                        sqrdistance = a + 2.0 * d + f
                    else:
                        s = numer / denom
                        t = 1 - s
                        sqrdistance = s * (a * s + b * t + 2.0 * d) + t * (b * s + c * t + 2.0 * e) + f

    # account for numerical round-off error
    if sqrdistance < 0:
        sqrdistance = 0

    dist = sqrdistance**0.5

    return dist


@numba.njit(cache=True)
def ray_triangle_intersect(ray_origin: np.ndarray, ray_direction: np.ndarray, triangle: np.ndarray) -> bool:
    """https://github.com/johnnovak/raytriangle-test/blob/master/python/perftest.py"""
    v0, v1, v2 = triangle[0], triangle[1], triangle[2]
    v0v1 = v1 - v0
    v0v2 = v2 - v0
    pvec = np.cross(ray_direction, v0v2)
    det = np.dot(v0v1, pvec)
    if det < 0.000001:
        return False
    invDet = 1.0 / det
    tvec = ray_origin - v0
    u = np.dot(tvec, pvec) * invDet
    if u < 0 or u > 1:
        return False
    qvec = np.cross(tvec, v0v1)
    v = np.dot(ray_direction, qvec) * invDet
    if v < 0 or u + v > 1:
        return False
    if np.dot(v0v2, qvec) * invDet < 0.000001:
        return False
    return True


@numba.njit(cache=True)
def segment_triangle_intersects(ray_origin: np.ndarray, ray_direction: np.ndarray, triangle: np.ndarray) -> bool:
    """Checks if a line segment (finite length) intersects with a triangle in 3D"""
    intersects1 = ray_triangle_intersect(ray_origin, ray_direction, triangle)
    intersects2 = ray_triangle_intersect((ray_origin + ray_direction), ray_direction * -1, triangle[::-1])
    return intersects1 and intersects2


@numba.njit(cache=True)
def segment_triangle_distance(p1: np.ndarray, p2: np.ndarray, triangle: np.ndarray) -> float:
    """Minimum distance between a line segment (finite length) and a triangle in 3D"""
    if segment_triangle_intersects(p1, p2, triangle):
        return 0.0
    distances = []
    for i, j in ((0, 1), (1, 2), (2, 0)):
        distances.append(segments_distance_3d(p1, p2, triangle[i], triangle[j]))
    return min(distances)



import pytest

PYTEST_BENCHMARK_PARAMS = {
    "min_time": 0.01,
    "max_time": 0.05,
    "min_rounds": 5,
    "warmup": True,
    "warmup_iterations": 1,
}


def get_pytest_benchmark(group: str):
    return pytest.mark.benchmark(group=group, **PYTEST_BENCHMARK_PARAMS)
